from tilings.assumptions import TrackingAssumption
from tilings.griddedperm import GriddedPerm
from tilings.tiling import Tiling

__version__ = "4.0.0"

__all__ = ["GriddedPerm", "Tiling", "TrackingAssumption"]



from typing import TYPE_CHECKING, Iterable, List, Optional, Tuple, Union

from logzero import logger

from comb_spec_searcher import StrategyPack
from comb_spec_searcher.strategies import (
    AbstractStrategy,
    Strategy,
    StrategyFactory,
    VerificationStrategy,
)
from permuta import Perm
from permuta.misc import DIR_EAST, DIR_NORTH, DIR_SOUTH, DIR_WEST, DIRS
from tilings import strategies as strat
from tilings.strategies.verification import BasisAwareVerificationStrategy

if TYPE_CHECKING:
    from tilings import Tiling

CSSstrategy = Union[Strategy, StrategyFactory, VerificationStrategy]


class TileScopePack(StrategyPack):
    # Method to add power to a pack
    # Pack are immutable, these methods return a new pack.

    def add_basis(self, basis: Iterable[Perm]) -> "TileScopePack":
        """
        Update the pack to add the basis being run to the verifications strategy
        that needs to be aware of it.
        """
        basis = tuple(basis)
        symmetry = bool(self.symmetries)

        def replace_list(
            strats: Tuple[Union[AbstractStrategy, StrategyFactory], ...]
        ) -> List[Union[AbstractStrategy, StrategyFactory]]:
            """Return a new list with the replaced 1x1 strat."""
            res: List[Union[AbstractStrategy, StrategyFactory]] = []
            for strategy in strats:
                if isinstance(strategy, BasisAwareVerificationStrategy):
                    if strategy.basis:
                        logger.warning("Basis changed in %s", strategy)
                    res.append(strategy.change_basis(basis, symmetry))
                elif isinstance(strategy, strat.SymmetriesFactory):
                    if strategy.basis:
                        logger.warning("Basis changed in %s", strategy)
                    res.append(strategy.change_basis(basis))
                elif hasattr(strategy, "change_basis"):
                    logger.warning("Basis changed in %s", strategy)
                    res.append(strategy.change_basis(basis))
                else:
                    res.append(strategy)
            return res

        return self.__class__(
            ver_strats=replace_list(self.ver_strats),
            inferral_strats=replace_list(self.inferral_strats),
            initial_strats=replace_list(self.initial_strats),
            expansion_strats=list(map(replace_list, self.expansion_strats)),
            name=self.name,
            symmetries=replace_list(self.symmetries),
            iterative=self.iterative,
        )

    def setup_subclass_verification(self, start_tiling: "Tiling") -> "TileScopePack":
        """
        If the subclass verification strategy already has a list of perms to check,
        we leave it alone. Otherwise we:
         - compute the cell basis for each cell
         - consider the permutations avoiding some cell basis
         - add to the list of perms to check any perm that avoids some cell basis and
           has length strictly smaller than the maximum length cell basis element.
        """

        def replace_list(
            strats: Tuple[Union[AbstractStrategy, StrategyFactory], ...]
        ) -> List[Union[AbstractStrategy, StrategyFactory]]:
            """
            Find subclass verification and alter its perms_to_check variable.
            """
            res: List[Union[AbstractStrategy, StrategyFactory]] = []
            for strategy in strats:
                if isinstance(strategy, strat.SubclassVerificationFactory):
                    printed_log = False
                    if strategy.perms_to_check is None:
                        new_perms_to_check = set()
                        cell_bases = set(
                            tuple(obs) for obs, _ in start_tiling.cell_basis().values()
                        )
                        max_length = (
                            max(max(len(b) for b in basis) for basis in cell_bases) - 1
                        )
                        perm_gen = Perm.up_to_length(max_length)
                        for perm in perm_gen:
                            if len(perm) == 0:
                                continue
                            if any(perm.avoids(*basis) for basis in cell_bases):
                                new_perms_to_check.add(perm)
                        res.append(strategy.change_perms(new_perms_to_check))
                        if start_tiling.dimensions == (1, 1):
                            logger.info(
                                "SubclassVerification set up to check the proper "
                                "principal subclasses of Av(%s)",
                                ", ".join(map(str, cell_bases.pop())),
                            )
                            printed_log = True
                    if not printed_log:
                        assert strategy.perms_to_check is not None
                        logger.info(
                            "SubclassVerification set up to check the subclasses: "
                            "Av(%s)",
                            "), Av(".join(map(str, strategy.perms_to_check)),
                        )
                else:
                    res.append(strategy)
            return res

        return self.__class__(
            ver_strats=replace_list(self.ver_strats),
            inferral_strats=replace_list(self.inferral_strats),
            initial_strats=replace_list(self.initial_strats),
            expansion_strats=list(map(replace_list, self.expansion_strats)),
            name=self.name,
            symmetries=self.symmetries,
            iterative=self.iterative,
        )

    def make_tracked(self):
        """Make a fusion pack tracked."""

        def replace_list(strats):
            """Return a new list with strats tracked."""
            res = []
            for strategy in strats:
                d = strategy.to_jsonable()
                if not d.get("tracked", True):
                    d["tracked"] = True
                    strategy = AbstractStrategy.from_dict(d)
                res.append(strategy)
            return res

        pack = self.__class__(
            ver_strats=replace_list(self.ver_strats),
            inferral_strats=replace_list(self.inferral_strats),
            initial_strats=replace_list(self.initial_strats),
            expansion_strats=list(map(replace_list, self.expansion_strats)),
            name=self.name.replace("untracked", "tracked"),
            symmetries=self.symmetries,
            iterative=self.iterative,
        )
        if all(
            not isinstance(strategy, strat.AddAssumptionFactory) for strategy in pack
        ):
            pack = pack.add_initial(strat.AddAssumptionFactory(), apply_first=True)
        if all(
            not isinstance(strategy, strat.RearrangeAssumptionFactory)
            for strategy in pack
        ):
            pack = pack.add_initial(
                strat.RearrangeAssumptionFactory(), apply_first=True
            )
        if strat.ComponentFusionFactory() in pack:
            if all(
                not isinstance(strategy, strat.DetectComponentsStrategy)
                for strategy in pack
            ):
                pack = pack.add_initial(
                    strat.DetectComponentsStrategy(ignore_parent=True)
                )
            if all(
                not isinstance(strategy, strat.ComponentVerificationStrategy)
                for strategy in pack
            ):
                pack = pack.add_verification(strat.ComponentVerificationStrategy())
        return pack

    def make_fusion(
        self,
        component: bool = False,
        tracked: bool = True,
        apply_first: bool = False,
        isolation_level: Optional[str] = None,
    ) -> "TileScopePack":
        """
        Create a new pack by adding fusion to the current pack.

        If component, it will add component fusion.
        If tracked, it will return the pack for finding a tracked tree.
        If apply_first, it will add fusion to the front of the initial strategies.
        """
        name = "tracked_" if tracked else "untracked_"
        name += "component_fusion" if component else "fusion"
        if isolation_level is not None:
            name += f"_{isolation_level}"
        if component:
            fusion_strat: StrategyFactory = strat.ComponentFusionFactory(
                tracked=tracked, isolation_level=isolation_level
            )
        else:
            fusion_strat = strat.FusionFactory(
                tracked=tracked, isolation_level=isolation_level
            )
        pack = self.add_initial(fusion_strat, name, apply_first=apply_first)
        if tracked:
            pack = pack.add_initial(strat.AddAssumptionFactory(), apply_first=True)
            if component:
                pack = pack.add_initial(
                    strat.DetectComponentsStrategy(ignore_parent=True), apply_first=True
                )
                pack = pack.add_verification(strat.ComponentVerificationStrategy())
            pack = pack.add_initial(
                strat.RearrangeAssumptionFactory(), apply_first=True
            )
        return pack

    def make_interleaving(
        self, tracked: bool = True, unions: bool = False
    ) -> "TileScopePack":
        """
        Return a new pack where the factor strategy is replaced with an
        interleaving factor strategy.

        If unions is set to True it will overwrite unions on the strategy.
        """

        def replace_list(strats):
            """Return a new list with the replaced tracked factor strategy."""
            res = []
            for strategy in strats:
                if isinstance(strategy, strat.FactorFactory):
                    d = strategy.to_jsonable()
                    d["interleaving"] = "all"
                    d["tracked"] = tracked
                    d["unions"] = d["unions"] or unions
                    res.append(AbstractStrategy.from_dict(d))
                else:
                    res.append(strategy)
            return res

        name = f"{self.name}_{'' if tracked else 'un'}tracked_interleaving"
        if unions:
            name += "_unions"

        pack = self.__class__(
            ver_strats=replace_list(self.ver_strats),
            inferral_strats=replace_list(self.inferral_strats),
            initial_strats=replace_list(self.initial_strats),
            expansion_strats=list(map(replace_list, self.expansion_strats)),
            name=name,
            symmetries=self.symmetries,
            iterative=self.iterative,
        )

        if tracked:
            pack = pack.add_initial(strat.AddAssumptionFactory(), apply_first=True)

        return pack

    def make_elementary(self) -> "TileScopePack":
        """
        Create a new pack by using only one by one and elementary
        verification.
        """
        if (
            strat.ElementaryVerificationStrategy() in self
            and strat.OneByOneVerificationStrategy() in self
            and len(self.ver_strats) == 2
            and self.iterative
        ):
            raise ValueError("The pack is already elementary.")
        pack = self.make_iterative()
        pack = pack.add_verification(strat.BasicVerificationStrategy(), replace=True)
        pack = pack.add_verification(strat.OneByOneVerificationStrategy())
        return pack.add_verification(
            strat.ElementaryVerificationStrategy(), "elementary"
        )

    def add_all_symmetry(self) -> "TileScopePack":
        """Create a new pack by turning on symmetry on the current pack."""
        if self.symmetries:
            raise ValueError("Symmetries already turned on.")
        return super().add_symmetry(strat.SymmetriesFactory(), "symmetries")

    def kitchen_sinkify(  # pylint: disable=R0912
        self,
        short_obs_len: int,
        obs_inferral_len: int,
        tracked: bool,
        level: int,
    ) -> "TileScopePack":
        """
        Create a new pack with the following added:
            Short Obs verification (unless short_obs_len = 0)
            No Root Cell verification
            Deflation
            Point and/or Assumption Jumping
            Generalized Monotone Sliding
            Free Cell Reduction
            Requirement corroboration
            Obstruction Inferral (unless obs_inferral_len = 0)
            Symmetries
            Point Pointing
            Unfusion
            Targeted Row/Col Placements when fusable
            Relax assumptions
        Will be made tracked or not, depending on preference.
        Note that nothing is done with positive / point corroboration, requirement
        corroboration.

        Different stratgies will be added at different levels
        Level 1: short obs, no root cell, symmetries,
                    obs inferral, interleaving factor without unions
        Level 2: deflation, point/assumption jumping, sliding, free cell reduction,
                    req corrob, targeted row/col placements, relax assumptions,
                    interleaving factor with unions
        Level 3: unfusion 1,1
        Level 4: unfusion 2,2, pointing mc=4, assumption mc=8
        Level 5: unfusion 4,4, pointing mc=6, assumption mc=8, requirement pt, mc=4
        """

        assert level in (
            1,
            2,
            3,
            4,
            5,
        ), "Level must be an int between 1 and 5 inclusive"

        ks_pack = self.__class__(
            ver_strats=self.ver_strats,
            inferral_strats=self.inferral_strats,
            initial_strats=self.initial_strats,
            expansion_strats=self.expansion_strats,
            name=self.name,
            symmetries=self.symmetries,
            iterative=self.iterative,
        )
        if short_obs_len > 0:
            ver_strats: List[CSSstrategy] = [
                strat.ShortObstructionVerificationStrategy(short_obs_len)
            ]
        else:
            ver_strats = []

        ver_strats += [
            strat.NoRootCellVerificationStrategy(),
        ]

        for strategy in ver_strats:
            try:
                ks_pack = ks_pack.add_verification(strategy)
            except ValueError:
                pass

        if level >= 2:
            initial_strats: List[CSSstrategy] = [
                strat.DeflationFactory(tracked),
                strat.AssumptionAndPointJumpingFactory(),
                strat.MonotoneSlidingFactory(),
                strat.CellReductionFactory(tracked),
                strat.RequirementCorroborationFactory(),
                strat.RelaxAssumptionFactory(),
            ]
            for strategy in initial_strats:
                try:
                    ks_pack = ks_pack.add_initial(strategy)
                except ValueError:
                    pass

        if obs_inferral_len > 0:
            inf_strats: List[CSSstrategy] = [
                strat.ObstructionInferralFactory(obs_inferral_len)
            ]
        else:
            inf_strats = []
        for strategy in inf_strats:
            try:
                ks_pack = ks_pack.add_inferral(strategy)
            except ValueError:
                pass

        ks_pack = ks_pack.make_interleaving(tracked=tracked, unions=level > 1)

        try:
            ks_pack = ks_pack.add_all_symmetry()
        except ValueError:
            pass

        if tracked:
            ks_pack = ks_pack.make_tracked()

        if level == 2:
            ks_pack.expansion_strats = ks_pack.expansion_strats + (
                (strat.FusableRowAndColumnPlacementFactory(),),
            )
        elif level == 3:
            ks_pack.expansion_strats = ks_pack.expansion_strats + (
                (
                    strat.UnfusionFactory(1, 1),
                    strat.FusableRowAndColumnPlacementFactory(),
                ),
            )
        elif level == 4:
            ks_pack.expansion_strats = ks_pack.expansion_strats + (
                (
                    strat.AssumptionPointingFactory(8),
                    strat.PointingStrategy(4),
                    strat.UnfusionFactory(2, 2),
                    strat.FusableRowAndColumnPlacementFactory(),
                ),
            )
        elif level == 5:
            ks_pack.expansion_strats = ks_pack.expansion_strats + (
                (
                    strat.AssumptionPointingFactory(8),
                    strat.RequirementPointingFactory(4),
                    strat.PointingStrategy(6),
                    strat.UnfusionFactory(4, 4),
                    strat.FusableRowAndColumnPlacementFactory(),
                ),
            )

        ks_pack.name += f"_kitchen_sink_level_{level}"

        return ks_pack

    # Creation of the base pack
    @classmethod
    def all_the_strategies(cls, length: int = 1) -> "TileScopePack":
        initial_strats: List[CSSstrategy] = [
            strat.FactorFactory(),
            strat.PointCorroborationFactory(),
        ]
        if length > 1:
            initial_strats.append(strat.RequirementCorroborationFactory())

        return TileScopePack(
            initial_strats=initial_strats,
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[
                [
                    strat.RequirementInsertionFactory(maxreqlen=length),
                    strat.AllPlacementsFactory(),
                ],
            ],
            name=f"all_the_strategies_{length}",
        )

    @classmethod
    def pattern_placements(
        cls, length: int = 1, partial: bool = False
    ) -> "TileScopePack":
        name = "".join(
            [
                f"length_{length}_" if length > 1 else "",
                "partial_" if partial else "",
                "pattern" if length > 1 else "pattern_point",
                "_placements",
            ]
        )

        expansion_strats: List[CSSstrategy] = [
            strat.CellInsertionFactory(maxreqlen=length),
        ]
        if length > 1:
            expansion_strats.append(strat.RequirementCorroborationFactory())

        return TileScopePack(
            initial_strats=[
                strat.FactorFactory(unions=True, ignore_parent=False),
                strat.PointCorroborationFactory(),
            ],
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[
                [strat.PatternPlacementFactory(partial=partial)],
                expansion_strats,
            ],
            name=name,
        )

    @classmethod
    def point_placements(
        cls, length: int = 1, partial: bool = False
    ) -> "TileScopePack":
        name = "".join(
            [
                f"length_{length}_" if length > 1 else "",
                "partial_" if partial else "",
                "point_placements",
            ]
        )

        initial_strats: List[CSSstrategy] = [
            strat.FactorFactory(),
            strat.PointCorroborationFactory(),
        ]
        if length > 1:
            initial_strats.append(strat.RequirementCorroborationFactory())

        return TileScopePack(
            initial_strats=initial_strats,
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[
                [
                    strat.CellInsertionFactory(maxreqlen=length),
                    strat.PatternPlacementFactory(partial=partial),
                ],
            ],
            name=name,
        )

    @classmethod
    def insertion_point_placements(cls, partial: bool = False) -> "TileScopePack":
        name = "insertion_"
        partial_str = "partial_" if partial else ""
        name += f"{partial_str}point_placements"
        return TileScopePack(
            initial_strats=[
                strat.FactorFactory(),
                strat.CellInsertionFactory(
                    maxreqlen=1, ignore_parent=True, one_cell_only=True
                ),
            ],
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[[strat.PatternPlacementFactory(partial=partial)]],
            name=name,
        )

    @classmethod
    def regular_insertion_encoding(cls, direction: int) -> "TileScopePack":
        """This pack finds insertion encodings."""
        if direction not in DIRS:
            raise ValueError(f"Must be direction in {DIRS}.")
        place_row = direction in (DIR_NORTH, DIR_SOUTH)
        place_col = not place_row
        direction_str = {
            DIR_EAST: "left",
            DIR_WEST: "right",
            DIR_NORTH: "bottom",
            DIR_SOUTH: "top",
        }
        name = f"regular_insertion_encoding_{direction_str[direction]}"
        return TileScopePack(
            initial_strats=[
                strat.FactorFactory(),
                strat.CellInsertionFactory(ignore_parent=True, one_cell_only=True),
            ],
            ver_strats=[strat.BasicVerificationStrategy()],
            inferral_strats=[],
            expansion_strats=[
                [
                    strat.RowAndColumnPlacementFactory(
                        place_col=place_col, place_row=place_row, dirs=[direction]
                    )
                ]
            ],
            name=name,
        )

    @classmethod
    def row_and_col_placements(
        cls, row_only: bool = False, col_only: bool = False, partial: bool = False
    ) -> "TileScopePack":
        if row_only and col_only:
            raise ValueError("Can't be row and col only.")
        place_row = not col_only
        place_col = not row_only
        both = place_col and place_row
        name = "".join(
            [
                "partial_" if partial else "",
                "row" if not col_only else "",
                "_and_" if both else "",
                "col" if not row_only else "",
                "_placements",
            ]
        )
        rowcol_strat = strat.RowAndColumnPlacementFactory(
            place_row=place_row, place_col=place_col, partial=partial
        )
        expansion_strats: List[List[CSSstrategy]] = [[rowcol_strat]]
        if partial:
            expansion_strats.append([strat.PatternPlacementFactory(point_only=True)])
        return TileScopePack(
            initial_strats=[
                strat.FactorFactory(),
                strat.PointCorroborationFactory(),
            ],
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=expansion_strats,
            name=name,
        )

    @classmethod
    def insertion_row_and_col_placements(
        cls, row_only: bool = False, col_only: bool = False, partial: bool = False
    ) -> "TileScopePack":
        pack = cls.row_and_col_placements(row_only, col_only, partial)
        pack.name = "insertion_" + pack.name
        pack = pack.add_initial(
            strat.CellInsertionFactory(
                maxreqlen=1, ignore_parent=True, one_cell_only=True
            )
        )
        return pack

    @classmethod
    def insertion_point_row_and_col_placements(
        cls, row_only: bool = False, col_only: bool = False, partial: bool = False
    ) -> "TileScopePack":
        pack = cls.insertion_row_and_col_placements(row_only, col_only, partial)
        if not partial:
            # if partial, then already added pattern placements
            pack.name = pack.name.replace("insertion", "insertion_point")
            pack.expansion_strats = pack.expansion_strats + (
                (strat.PatternPlacementFactory(partial=False),),
            )
        return pack

    @classmethod
    def only_root_placements(
        cls,
        length: int = 3,
        max_num_req: Optional[int] = 1,
        max_placement_rules_per_req: Optional[int] = None,
        partial: bool = False,
    ) -> "TileScopePack":
        partial_str = "partial_" if partial else ""
        if max_num_req is not None:
            name = (
                f"only_length_{length}_{max_num_req}_reqs_root_{partial_str}placements"
            )
        else:
            name = f"only_length_{length}_root_{partial_str}placements"
        placement_factory = strat.RequirementPlacementFactory(
            max_rules_per_req=max_placement_rules_per_req, partial=partial
        )
        return TileScopePack(
            initial_strats=[
                strat.RootInsertionFactory(maxreqlen=length, max_num_req=max_num_req),
                strat.FactorFactory(unions=True, ignore_parent=False, workable=False),
                strat.PointCorroborationFactory(),
            ],
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[[placement_factory]],
            name=name,
        )

    @classmethod
    def requirement_placements(
        cls, length: int = 2, partial: bool = False
    ) -> "TileScopePack":
        name = "".join(
            [
                f"length_{length}_" if length != 2 else "",
                "partial_" if partial else "",
                "requirement_placements",
            ]
        )

        initial_strats: List[CSSstrategy] = [
            strat.FactorFactory(),
            strat.PointCorroborationFactory(),
        ]
        if length > 1:
            initial_strats.append(strat.RequirementCorroborationFactory())

        return TileScopePack(
            initial_strats=initial_strats,
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[
                [
                    strat.RequirementInsertionFactory(maxreqlen=length),
                    strat.PatternPlacementFactory(partial=partial),
                ],
            ],
            name=name,
        )

    @classmethod
    def subobstruction_placements(cls, partial: bool = False) -> "TileScopePack":
        name = "partial_" if partial else ""
        name += "subobstruction_placements"
        return TileScopePack(
            initial_strats=[
                strat.FactorFactory(),
                strat.PointCorroborationFactory(),
                strat.RequirementCorroborationFactory(),
            ],
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[
                [
                    strat.SubobstructionInsertionFactory(),
                    strat.PatternPlacementFactory(partial=partial),
                ],
            ],
            name=name,
        )

    @classmethod
    def basis_pattern_insertions(cls, partial: bool = False) -> "TileScopePack":
        name = "partial_" if partial else ""
        name += "basis_pattern_insertions"
        return TileScopePack(
            initial_strats=[
                strat.FactorFactory(),
                strat.PointCorroborationFactory(),
                strat.RequirementCorroborationFactory(),
            ],
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[
                [
                    strat.BasisPatternInsertionFactory(),
                    strat.PatternPlacementFactory(partial=partial),
                ],
            ],
            name=name,
        )

    @classmethod
    def point_and_row_and_col_placements(
        cls,
        length: int = 1,
        row_only: bool = False,
        col_only: bool = False,
        partial: bool = False,
    ) -> "TileScopePack":
        if row_only and col_only:
            raise ValueError("Can't be row and col only.")
        place_row = not col_only
        place_col = not row_only
        both = place_col and place_row
        name = "".join(
            [
                f"length_{length}_" if length > 1 else "",
                "partial_" if partial else "",
                "point_and_",
                "row" if not col_only else "",
                "_and_" if both else "",
                "col" if not row_only else "",
                "_placements",
            ]
        )
        rowcol_strat = strat.RowAndColumnPlacementFactory(
            place_row=place_row, place_col=place_col, partial=partial
        )

        initial_strats: List[CSSstrategy] = [
            strat.FactorFactory(),
            strat.PointCorroborationFactory(),
        ]
        if length > 1:
            initial_strats.append(strat.RequirementCorroborationFactory())

        return TileScopePack(
            initial_strats=initial_strats,
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[
                [
                    strat.CellInsertionFactory(maxreqlen=length),
                    strat.PatternPlacementFactory(partial=partial),
                    rowcol_strat,
                ],
            ],
            name=name,
        )

    @classmethod
    def requirement_and_row_and_col_placements(
        cls,
        length: int = 1,
        row_only: bool = False,
        col_only: bool = False,
        partial: bool = False,
    ) -> "TileScopePack":
        if row_only and col_only:
            raise ValueError("Can't be row and col only.")
        place_row = not col_only
        place_col = not row_only
        both = place_col and place_row
        name = "".join(
            [
                f"length_{length}_" if length > 1 else "",
                "partial_" if partial else "",
                "requirement_and_",
                "row" if not col_only else "",
                "_and_" if both else "",
                "col" if not row_only else "",
                "_placements",
            ]
        )
        rowcol_strat = strat.RowAndColumnPlacementFactory(
            place_row=place_row, place_col=place_col, partial=partial
        )

        initial_strats: List[CSSstrategy] = [
            strat.FactorFactory(),
            strat.PointCorroborationFactory(),
        ]
        if length > 1:
            initial_strats.append(strat.RequirementCorroborationFactory())

        return TileScopePack(
            initial_strats=initial_strats,
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
                strat.LocallyFactorableVerificationStrategy(),
            ],
            inferral_strats=[
                strat.RowColumnSeparationStrategy(),
                strat.ObstructionTransitivityFactory(),
            ],
            expansion_strats=[
                [
                    strat.RequirementInsertionFactory(maxreqlen=length),
                    strat.PatternPlacementFactory(partial=partial),
                    rowcol_strat,
                ],
            ],
            name=name,
        )

    @classmethod
    def cell_insertions(cls, length: int):
        return TileScopePack(
            initial_strats=[],
            ver_strats=[
                strat.BasicVerificationStrategy(),
                strat.InsertionEncodingVerificationStrategy(),
                strat.OneByOneVerificationStrategy(),
            ],
            inferral_strats=[],
            expansion_strats=[[strat.CellInsertionFactory(maxreqlen=length)]],
            name=f"length_{length}_cell_insertions",
        )



from collections import defaultdict
from itertools import chain, combinations
from typing import TYPE_CHECKING, Dict, Iterable, Iterator, List, Optional, Set, Tuple

from permuta.misc import UnionFind
from tilings import GriddedPerm
from tilings.assumptions import ComponentAssumption, TrackingAssumption
from tilings.misc import partitions_iterator

if TYPE_CHECKING:
    from tilings import Tiling


Cell = Tuple[int, int]
ReqList = Tuple[GriddedPerm, ...]


class Factor:
    """
    Algorithm to compute the factorisation of a tiling.

    Two active cells are in the same factor if they are in the same row
    or column, or they share an obstruction or a requirement.

    If using tracking assumptions, then two cells will also be in the same
    factor if they are covered by the same assumption.
    """

    def __init__(self, tiling: "Tiling") -> None:
        self._tiling = tiling
        self._active_cells = tiling.active_cells
        nrow = tiling.dimensions[1]
        ncol = tiling.dimensions[0]
        self._cell_unionfind = UnionFind(nrow * ncol)
        self._components: Optional[Tuple[Set[Cell], ...]] = None
        self._factors_obs_and_reqs: Optional[
            List[
                Tuple[
                    Tuple[GriddedPerm, ...],
                    Tuple[ReqList, ...],
                    Tuple[TrackingAssumption, ...],
                ]
            ]
        ] = None

    def _cell_to_int(self, cell: Cell) -> int:
        nrow = self._tiling.dimensions[1]
        return cell[0] * nrow + cell[1]

    def _int_to_cell(self, i: int) -> Cell:
        nrow = self._tiling.dimensions[1]
        return (i // nrow, i % nrow)

    def _get_cell_representative(self, cell: Cell) -> Cell:
        """
        Return the representative of a cell in the union find.
        """
        i = self._cell_to_int(cell)
        return self._cell_unionfind.find(i)  # type: ignore

    def _unite_cells(self, cells: Iterable[Cell]) -> None:
        """
        Put all the cells of `cells` in the same component of the UnionFind.
        """
        cell_iterator = iter(cells)
        try:
            c1 = next(cell_iterator)
        except StopIteration:
            return
        c1_int = self._cell_to_int(c1)
        for c2 in cell_iterator:
            c2_int = self._cell_to_int(c2)
            self._cell_unionfind.unite(c1_int, c2_int)

    def _unite_assumptions(self) -> None:
        """
        For each TrackingAssumption unite all the positions of the gridded perms.
        """
        for assumption in self._tiling.assumptions:
            if isinstance(assumption, ComponentAssumption):
                for cells in assumption.cell_decomposition(self._tiling):
                    self._unite_cells(cells)
            else:
                for gp in assumption.gps:
                    self._unite_cells(gp.pos)

    def _unite_obstructions(self) -> None:
        """
        For each obstruction unite all the position of the obstruction.
        """
        for ob in self._tiling.obstructions:
            self._unite_cells(ob.pos)

    def _unite_requirements(self) -> None:
        """
        For each requirement unite all the cell in all the requirements of the
        list.
        """
        for req_list in self._tiling.requirements:
            req_cells = chain.from_iterable(req.pos for req in req_list)
            self._unite_cells(req_cells)

    @staticmethod
    def _same_row_or_col(cell1: Cell, cell2: Cell) -> bool:
        """
        Test if `cell1` and `cell2` or in the same row or columns
        """
        return cell1[0] == cell2[0] or cell1[1] == cell2[1]

    def _unite_rows_and_cols(self) -> None:
        """
        Unite all the active cell that are on the same row or column.
        """
        cell_pair_to_unite = (
            c
            for c in combinations(self._active_cells, r=2)
            if self._same_row_or_col(c[0], c[1])
        )
        for c1, c2 in cell_pair_to_unite:
            self._unite_cells((c1, c2))

    def _unite_all(self) -> None:
        """
        Unite all the cells that share an obstruction, a requirement list,
        a row or a column.
        """
        self._unite_obstructions()
        self._unite_requirements()
        self._unite_assumptions()
        self._unite_rows_and_cols()

    def get_components(self) -> Tuple[Set[Cell], ...]:
        """
        Returns the tuple of all the components. Each component is set of
        cells.
        """
        if self._components is not None:
            return self._components
        self._unite_all()
        all_components: Dict[Cell, Set[Cell]] = defaultdict(set)
        for cell in self._active_cells:
            rep = self._get_cell_representative(cell)
            all_components[rep].add(cell)
        self._components = tuple(all_components.values())
        return self._components

    def _get_factors_obs_and_reqs(
        self,
    ) -> List[
        Tuple[
            Tuple[GriddedPerm, ...], Tuple[ReqList, ...], Tuple[TrackingAssumption, ...]
        ],
    ]:
        """
        Returns a list of all the irreducible factors of the tiling.
        Each factor is a tuple (obstructions, requirements)
        """
        if self._factors_obs_and_reqs is not None:
            return self._factors_obs_and_reqs
        if self._tiling.is_empty():
            return [((GriddedPerm((), []),), tuple(), tuple())]
        factors = []
        for component in self.get_components():
            obstructions = tuple(
                ob for ob in self._tiling.obstructions if ob.pos[0] in component
            )
            requirements = tuple(
                req for req in self._tiling.requirements if req[0].pos[0] in component
            )
            # TODO: consider skew/sum assumptions
            assumptions = tuple(
                ass.__class__(gp for gp in ass.gps if gp.pos[0] in component)
                for ass in self._tiling.assumptions
            )
            factors.append(
                (
                    obstructions,
                    requirements,
                    tuple(set(ass for ass in assumptions if ass.gps)),
                )
            )
        self._factors_obs_and_reqs = factors
        return self._factors_obs_and_reqs

    def factorable(self) -> bool:
        """
        Returns `True` if the tiling has more than one factor.
        """
        return len(self.get_components()) > 1

    def factors(self) -> Tuple["Tiling", ...]:
        """
        Returns all the irreducible factors of the tiling.
        """
        return tuple(
            self._tiling.__class__(
                obstructions=f[0],
                requirements=f[1],
                assumptions=tuple(sorted(f[2])),
                simplify=False,
            )
            for f in self._get_factors_obs_and_reqs()
        )

    def reducible_factorisations(self) -> Iterator[Tuple["Tiling", ...]]:
        """
        Iterator over all reducible factorisation that can be obtained by
        grouping of irreducible factors.

        Each factorisation is a list of tiling.

        For example if T = T1 x T2 x T3 then (T1 x T3) x T2 is a possible
        reducible factorisation.
        """
        min_comp = self._get_factors_obs_and_reqs()
        for partition in partitions_iterator(min_comp):
            factors = []
            for part in partition:
                obstructions, requirements, assumptions = zip(*part)
                factors.append(
                    self._tiling.__class__(
                        obstructions=chain(*obstructions),
                        requirements=chain(*requirements),
                        assumptions=tuple(sorted(chain(*assumptions))),
                        simplify=False,
                    )
                )
            yield tuple(factors)


class FactorWithMonotoneInterleaving(Factor):
    """
    Algorithm to compute the factorisation with monotone interleaving of a
    tiling.

    Two active cells are in the same factor if they share an obstruction or
    a requirement or if they are on the same row or column and are both
    non-monotone.
    """

    def _unite_rows_and_cols(self) -> None:
        """
        Unite all active cell that are on the same row or column if both of
        them are not monotone.

        Override `Factor._unite_rows_and_cols`.
        """
        cell_pair_to_unite = (
            c
            for c in combinations(self._active_cells, r=2)
            if (
                self._same_row_or_col(c[0], c[1])
                and not self._tiling.is_monotone_cell(c[0])
                and not self._tiling.is_monotone_cell(c[1])
            )
        )
        for c1, c2 in cell_pair_to_unite:
            self._unite_cells((c1, c2))


class FactorWithInterleaving(Factor):
    """
    Algorithm to compute the factorisation with interleaving of a tiling.

    Two non-empty cells are in the same factor if they share an obstruction or
    a requirement.
    """

    def _unite_rows_and_cols(self) -> None:
        """
        Override the `Factor._unite_rows_and_cols` to do nothing since
        interleaving is allowed on row and column.
        """

    def _unite_all(self) -> None:
        """
        Unite all the cells that share an obstruction or a requirement list.
        """
        self._unite_assumptions()
        self._unite_obstructions()
        self._unite_requirements()



from .component import ComponentFusionFactory, ComponentFusionStrategy
from .constructor import FusionConstructor
from .fusion import FusionFactory, FusionRule, FusionStrategy

__all__ = [
    "ComponentFusionFactory",
    "ComponentFusionStrategy",
    "FusionFactory",
    "FusionStrategy",
    "FusionRule",
    "FusionConstructor",
]



"""
The row and columns separation algorithm.

The main algorithm for the separation is contained in the class
`_RowColSeparationSingleApplication`. The new separated tiling have new
length 2 crossing obstructions that would allow more separation.
The role of the class `RowColSeparation` is to make sure that row columns
separation is idempotent by applying the core algorithm until it stabilises.
"""
import heapq
from itertools import combinations, product
from typing import TYPE_CHECKING, Dict, List, Tuple

from tilings import GriddedPerm

if TYPE_CHECKING:
    from tilings import Tiling

Cell = Tuple[int, int]


class Graph:
    """
    A weighted directed graph implemented with an adjacency matrix.

    The graph is made such that it is easy to merge to vertices. Merging
    vertices collapse to vertices together such that
        - The weight of the new vertex is the sum of the weights
        - The weight of the edges is the sum of the weight of the edges that
        went to any of the merged vertices before the merge.


    The graph supports 2 operations
        - `reduce`: who merge two vertices that were not connected by an edges
        and repeat as long as possible.
        - `break_cycle_in_all_ways`: Take a cycle in the graph and return a
        copy of the graph with a removed edges for each edges in the cycle.

    Moreover, one can also ask:
        - if the graph is acyclic with `is_acyclic`
        - for a cycle of the graph with `find_cycle`
        - For the vertex order implied by a reduced acyclic graph
    """

    def __init__(self, vertices, matrix=None):
        self._vertex_labels = [set([v]) for v in vertices]
        self._vertex_weights = [1 for _ in self._vertex_labels]
        self._matrix = matrix
        assert len(matrix) == len(self._vertex_labels)
        assert all(len(row) == len(self._matrix) for row in matrix)
        self._reduced = False
        self._is_acyclic = False

    @property
    def num_vertices(self):
        """
        The number of vertices of the graph
        """
        return len(self._vertex_weights)

    def _merge_vertices(self, v1, v2):
        """
        Merge the two vertices.

        Vertex and edges are merged and the weight are added. Then edges with a
        weight that is to small are discarded.
        """
        v2_label = self._vertex_labels.pop(v2)
        self._vertex_labels[v1].update(v2_label)
        v2_weight = self._vertex_weights.pop(v2)
        self._vertex_weights[v1] += v2_weight
        self._add_matrix_rows(v1, v2)
        self._add_matrix_columns(v1, v2)
        self._trim_edges(v1)

    def reduce(self):
        if self._reduced:
            return
        non_edge = self.find_non_edge()
        while non_edge:
            self._merge_vertices(non_edge[0], non_edge[1])
            non_edge = self.find_non_edge()
        self._reduced = True

    def find_non_edge(self):
        """
        Return a non-edge of the graph.

        A non edges is a pair of vertices `(v1, v2)` such that neither
        `(v1, v2)` or `(v2, v1)` is an edge in the graph.
        """
        for v1, v2 in combinations(range(self.num_vertices), 2):
            if not self._is_edge(v1, v2) and not self._is_edge(v2, v1):
                return (v1, v2)

    def is_acyclic(self):
        """
        Check if the graph is acyclic.

        To perform that check, the graph must first be reduced with the
        `reduce` method.
        """
        assert self._reduced, "Graph must first be reduced"
        if self._is_acyclic or self.num_vertices == 0:
            return True
        return self.find_cycle() is None

    def find_cycle(self):
        """
        Return the edges of a cycle of the graphs. The graphs first need to be
        reduced

        If a cycle of length 3 is return it means that no cycle of length 2
        exist.

        If the graph is acyclic, returns None.

        NOTE:

            One can prove that if a reduced graph is not acyclic it has either
            a cycle of length 2 or 3.
        """
        assert self._reduced, "Graph must first be reduced"
        for v1, v2 in combinations(range(self.num_vertices), 2):
            if self._is_edge(v1, v2) and self._is_edge(v2, v1):
                return ((v1, v2), (v2, v1))
        for v1, v2, v3 in combinations(range(self.num_vertices), 3):
            cycle = self._length3_cycle(v1, v2, v3)
            if cycle:
                return cycle
        self._is_acyclic = True
        return None

    def break_cycle_in_all_ways(self, edges):
        """
        Generator over Graph object obtained by removing one edge of the
        `edges` iterator.
        """
        # pylint: disable=protected-access
        for e in edges:
            new_graph = Graph.__new__(Graph)
            new_graph._vertex_labels = [vl.copy() for vl in self._vertex_labels]
            new_graph._vertex_weights = self._vertex_weights.copy()
            new_graph._matrix = [row.copy() for row in self._matrix]
            new_graph._matrix[e[0]][e[1]] = 0
            new_graph._reduced = False
            new_graph._is_acyclic = False
            yield new_graph

    def vertex_order(self):
        """
        Return the order of the vertex in a reduced acyclic graph.

        A reduced acyclic graph is an acyclic orientation of a complete graph.
        There it equivalent to an ordering of its vertices.

        To compute the vertex order, the graph must be reduced and acyclic.
        """
        assert self._reduced, "Graph must first be reduced"
        assert self.is_acyclic(), "Graph must be acyclic"
        vert_num_parent = [row.count(0) for row in self._matrix]
        return [p[1] for p in sorted(zip(vert_num_parent, self._vertex_labels))]

    def _add_matrix_rows(self, row1_idx, row2_idx):
        """
        Deletes row 2 from the graph matrix and change row 1 to
        the sum of both row.
        """
        assert row1_idx != row2_idx
        row1 = self._matrix[row1_idx]
        row2 = self._matrix.pop(row2_idx)
        self._matrix[row1_idx] = list(map(sum, zip(row1, row2)))

    def _add_matrix_columns(self, col1_idx, col2_idx):
        """
        Deletes column 2 from the graph matrix and change column 1 to
        the sum of both column.
        """
        assert col1_idx != col2_idx
        for row in self._matrix:
            c2_value = row.pop(col2_idx)
            row[col1_idx] += c2_value

    def _trim_edges(self, vertex):
        """
        Remove all the edges that touch vertex that that have a weight which is
        too small.

        The weight of a vertex is too small if it is smaller than the product
        of the weights of the two vertex it connects.
        """
        v1 = vertex
        v1_weight = self._vertex_weights[v1]
        for v2 in range(self.num_vertices):
            v2_weight = self._vertex_weights[v2]
            weight_prod = v1_weight * v2_weight
            self._delete_edge_if_small(v1, v2, weight_prod)
            self._delete_edge_if_small(v2, v1, weight_prod)

    def _delete_edge_if_small(self, head, tail, cap):
        """
        Delete the edges that goes from head to tail if its weight is lower
        than the cap.
        """
        weight = self._matrix[head][tail]
        if weight < cap:
            self._matrix[head][tail] = 0

    def _is_edge(self, v1, v2):
        return self._matrix[v1][v2] != 0

    def _length3_cycle(self, v1, v2, v3):
        """
        Return the edges of a length 3 cycle containing the three vertices if
        such a cycle exist. Otherwise return None
        """

        def is_cycle(edges):
            return all(self._is_edge(*e) for e in edges)

        orientation1 = ((v1, v2), (v2, v3), (v3, v1))
        if is_cycle(orientation1):
            return orientation1
        orientation2 = ((v1, v3), (v3, v2), (v2, v1))
        if is_cycle(orientation2):
            return orientation2

    def __repr__(self):
        s = f"Graph over the vertices {self._vertex_labels}\n"
        s += f"Vertex weight is {self._vertex_weights}\n"
        for row in self._matrix:
            s += f"{row}\n"
        return s

    def __lt__(self, other):
        """
        A graph is 'smaller if it as more vertices.
        Useful for the priority queue
        """
        return self.num_vertices > other.num_vertices

    def __le__(self, other):
        """
        A graph is 'smaller if it as more vertices.
        Useful for the priority queue
        """
        return self.num_vertices >= other.num_vertices


class _RowColSeparationSingleApplication:
    """
    Make the row separation of the tiling.
    """

    def __init__(self, tiling):
        self._tiling = tiling
        self._active_cells = tuple(sorted(tiling.active_cells))
        self._ineq_matrices = None
        self._max_row_order = None
        self._max_col_order = None

    def cell_at_idx(self, idx):
        """Return the cell at index `idx`."""
        return self._active_cells[idx]

    def cell_idx(self, cell):
        """Return the index of the cell"""
        return self._active_cells.index(cell)

    def _basic_matrix(self, row):
        """
        Compute the basic matrix of inequalities based only on difference in
        row and columns. If `row` is True return the matrix for the row,
        otherwise return if for the columns.
        """
        idx = 1 if row else 0
        m = []
        for c1 in self._active_cells:
            row = [1 if c1[idx] < c2[idx] else 0 for c2 in self._active_cells]
            m.append(row)
        return m

    @staticmethod
    def _row_cell_order(ob):
        """
        Return the order of the two cells of a length 2 obstruction localized
        in a row.


        A cell `c1` is smaller than cell `c2` if all the points in `c1` are
        lower than all the point in `c2`.

        OUTPUT:
            A tuple (smaller_cell, bigger_cell)
        """
        assert len(ob) == 2, "Obstruction must be of length 2"
        c1, c2 = ob.pos
        assert c1[1] == c2[1], "Obstruction not a row obstruction"
        assert c1[0] != c2[0], "Obstruction is single cell"
        if ob.patt[0] == 0:
            return c2, c1
        return c1, c2

    @staticmethod
    def _col_cell_order(ob):
        """
        Return the order of the two cells of a length 2 obstruction.

        A cell `c1` is smaller than a cell `c2` if all the points in `c1` must
        be on the left all points in `c2`.

        OUTPUT:
            A tuple (smaller_cell, bigger_cell)
        """
        assert len(ob) == 2, "Obstruction must be of length 2"
        c1, c2 = ob.pos
        assert c1[0] == c2[0], "Obstruction not a col obstruction"
        assert not c1[1] == c2[1], "Obstruction is single cell"
        return c2, c1

    def _add_ineq(self, ineq, matrix):
        """
        Add an inequalities to the matrix.

        The inequalities must a tuple (smaller_cell, bigger_cell).
        """
        small_c, big_c = ineq
        matrix[self.cell_idx(small_c)][self.cell_idx(big_c)] = 1

    def _complete_ineq_matrices(self):
        """
        Return the matrices of inequalities between the cells.

        OUTPUT:
            tuple `(row_matrix, col_matrix)`
        """
        if self._ineq_matrices is not None:
            return self._ineq_matrices
        row_m = self._basic_matrix(row=True)
        col_m = self._basic_matrix(row=False)
        filtered_obs = (
            ob
            for ob in self._tiling.obstructions
            if len(ob) == 2 and not ob.is_single_cell()
        )
        for ob in filtered_obs:
            c1, c2 = ob.pos
            if c1[1] == c2[1]:
                ineq = self._row_cell_order(ob)
                self._add_ineq(ineq, row_m)
            elif c1[0] == c2[0]:
                ineq = self._col_cell_order(ob)
                self._add_ineq(ineq, col_m)
        self._ineq_matrices = row_m, col_m
        return self._ineq_matrices

    def row_ineq_graph(self):
        return Graph(self._active_cells, self._complete_ineq_matrices()[0])

    def col_ineq_graph(self):
        return Graph(self._active_cells, self._complete_ineq_matrices()[1])

    @staticmethod
    def _all_order(graph, only_max=False):
        """
        Generator of ordering of the active cells.

        One can get only the maximal separation by setting `only_max` to
        `True`.

        The order are yielded in decreasing order of size.
        """
        max_sep_seen = 0
        graph.reduce()
        heap = [graph]
        while heap and (not only_max or max_sep_seen <= graph.num_vertices):
            graph = heapq.heappop(heap)
            cycle = graph.find_cycle()
            if cycle is None:
                yield graph.vertex_order()
            else:
                for g in graph.break_cycle_in_all_ways(cycle):
                    g.reduce()
                    heapq.heappush(heap, g)

    @staticmethod
    def _maximal_order(graph):
        """Returns a order that maximise separation."""
        return next(_RowColSeparationSingleApplication._all_order(graph))

    def _separates_tiling(self, row_order, col_order):
        cell_map = self._get_cell_map(row_order, col_order)
        obs = self.map_obstructions(cell_map)
        reqs = self.map_requirements(cell_map)
        ass = self.map_assumptions(cell_map)
        return self._tiling.__class__(
            obstructions=obs, requirements=reqs, assumptions=ass
        )

    @staticmethod
    def _get_cell_map(row_order, col_order) -> Dict[Cell, Cell]:
        """
        Return the position of the according to the given row_order and
        col_order.

        This method does not account for any cleaning occuring in the initializer. For
        the complete cell map use `get_cell_map`.
        """
        cell_map: Dict[Cell, Cell] = {}
        row_map: Dict[Cell, int] = {}
        for i, row in enumerate(row_order):
            for cell in row:
                row_map[cell] = i
        for i, col in enumerate(col_order):
            for cell in col:
                cell_map[cell] = (i, row_map[cell])
        return cell_map

    def map_obstructions(self, cell_map):
        """Map the obstruction of a tiling according to the cell map."""
        non_point_obs = (ob for ob in self._tiling.obstructions if len(ob) > 1)
        for ob in non_point_obs:
            ob = self._map_gridded_perm(cell_map, ob)
            if not ob.contradictory():
                yield ob

    def map_requirements(self, cell_map):
        """Map the requirements of a tiling according to the cell map."""
        for req_list in self._tiling.requirements:
            yield [self._map_gridded_perm(cell_map, req) for req in req_list]

    def map_assumptions(self, cell_map):
        """Map the assumptions of a tiling according to the cell map."""
        for ass in self._tiling.assumptions:
            gps: List[GriddedPerm] = []
            for gp in ass.gps:
                mapped_gp = self._map_gridded_perm(cell_map, gp)
                if not mapped_gp.contradictory():
                    gps.append(mapped_gp)
            yield ass.__class__(gps)

    @property
    def max_row_order(self):
        """A maximal order on the rows."""
        if self._max_row_order is not None:
            return self._max_row_order
        self._max_row_order = self._maximal_order(self.row_ineq_graph())
        return self._max_row_order

    @property
    def max_col_order(self):
        """A maximal order on the columns."""
        if self._max_col_order is not None:
            return self._max_col_order
        self._max_col_order = self._maximal_order(self.col_ineq_graph())
        return self._max_col_order

    @staticmethod
    def _map_gridded_perm(cell_map, gp):
        """
        Transform a gridded perm by mapping the position of the gridded perm
        according to the cell_map
        """
        pos = (cell_map[p] for p in gp.pos)
        gp = gp.__class__(gp.patt, pos)
        return gp

    def separable(self):
        """
        Test if the tiling is separable.

        A tiling is not separable the separation does not creates any new row
        or column.
        """
        ncol, nrow = self._tiling.dimensions
        return len(self.max_row_order) > nrow or len(self.max_col_order) > ncol

    def separated_tiling(self):
        """
        Return the one the possible maximal separation of the tiling.
        """
        return self._separates_tiling(self.max_row_order, self.max_col_order)

    def get_cell_map(self) -> Dict[Cell, Cell]:
        """
        Return the position of the according to the given row_order and
        col_order. This accounts for any cleaning happening inside tiling initializer.

        This is the cell map for the separated tiling returned by `separated_tiling`.
        """
        sep_tiling = self.separated_tiling()
        row_order = self.max_row_order
        col_order = self.max_col_order
        sep_cell_map = self._get_cell_map(row_order, col_order)
        init_cell_map = sep_tiling.forward_map
        res: Dict[Cell, Cell] = {}
        for cell in self._tiling.active_cells:
            mid_cell = sep_cell_map[cell]
            # If the cell is not in the init map it is an empty cell
            if init_cell_map.is_mappable_cell(mid_cell):
                final_cell = init_cell_map.map_cell(mid_cell)
                res[cell] = final_cell
        return res

    def all_separated_tiling(self, only_max=False):
        """
        Generator over all the possibles separation of the tiling.

        One can only get the maximal one by setting `only_max=True`.

        NOTE: The same tiling might be returned many times.
        """
        orders = product(
            self._all_order(self.row_ineq_graph(), only_max=only_max),
            self._all_order(self.col_ineq_graph(), only_max=only_max),
        )
        for row_order, col_order in orders:
            yield self._separates_tiling(row_order, col_order)


class RowColSeparation:
    """
    This is basically a simple wrapper around
    `_RowColSeparationSingleApplication` to ensure the separation is
    idempotent.

    It applies the row columns separation until it does not change the tiling.
    """

    def __init__(self, tiling: "Tiling") -> None:
        self._tiling = tiling
        self._separated_tilings: List["Tiling"] = []
        separation_algo = _RowColSeparationSingleApplication(self._tiling)
        while separation_algo.separable():
            new_sep = separation_algo.separated_tiling()
            self._separated_tilings.append(new_sep)
            separation_algo = _RowColSeparationSingleApplication(new_sep)

    def separable(self) -> bool:
        """
        Test if the tiling is separable.

        A tiling is not separable the separation does not creates any new row
        or column.
        """
        return bool(self._separated_tilings)

    def separated_tiling(self) -> "Tiling":
        """
        Return the one the possible maximal separation of the tiling.
        """
        if not self._separated_tilings:
            return self._tiling
        return self._separated_tilings[-1]

    def get_cell_map(self) -> Dict[Cell, Cell]:
        """
        Return the cell map obtained by applying the algorithm until no change.
        """
        separation_algo = _RowColSeparationSingleApplication(self._tiling)
        cell_maps = []
        while separation_algo.separable():
            cell_map = separation_algo.get_cell_map()
            cell_maps.append(cell_map)
            new_sep = separation_algo.separated_tiling()
            separation_algo = _RowColSeparationSingleApplication(new_sep)
        res = {cell: cell for cell in self._tiling.active_cells}
        for cell_map in cell_maps:
            for cell, mapped_cell in tuple(res.items()):
                if mapped_cell in cell_map:
                    res[cell] = cell_map[mapped_cell]
                else:
                    res.pop(cell)
        return res



"""
The directionless point placement strategy that is counted
by the 'pointing' constructor.
"""
from collections import Counter
from itertools import product
from typing import (
    Callable,
    Dict,
    FrozenSet,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

from comb_spec_searcher import Strategy
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.strategies.constructor.disjoint import DisjointUnion
from comb_spec_searcher.strategies.rule import Rule
from comb_spec_searcher.strategies.strategy import StrategyFactory
from comb_spec_searcher.typing import CombinatorialClassType, SubTerms, Terms
from permuta.misc import DIR_NONE
from tilings import GriddedPerm, Tiling
from tilings.algorithms import RequirementPlacement
from tilings.assumptions import ComponentAssumption, TrackingAssumption
from tilings.strategies.assumption_insertion import AddAssumptionsStrategy
from tilings.strategies.obstruction_inferral import ObstructionInferralStrategy
from tilings.tiling import Cell

from .unfusion import DivideByN, ReverseDivideByN


class PointingStrategy(Strategy[Tiling, GriddedPerm]):
    def __init__(
        self,
        max_cells: Optional[int] = 4,
        ignore_parent: bool = False,
        inferrable: bool = True,
        possibly_empty: bool = True,
        workable: bool = True,
    ) -> None:
        self.max_cells = max_cells
        super().__init__(ignore_parent, inferrable, possibly_empty, workable)

    def can_be_equivalent(self) -> bool:
        return False

    def is_two_way(self, comb_class: Tiling) -> bool:
        return True

    def is_reversible(self, comb_class: Tiling) -> bool:
        return True

    def shifts(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]],
    ) -> Tuple[int, ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        assert children is not None
        return tuple(0 for _ in children)

    @staticmethod
    def already_placed_cells(comb_class: Tiling) -> FrozenSet[Cell]:
        return frozenset(
            cell
            for cell in comb_class.point_cells
            if comb_class.only_cell_in_row_and_col(cell)
        )

    def cells_to_place(self, comb_class: Tiling) -> FrozenSet[Cell]:
        return comb_class.active_cells - self.already_placed_cells(comb_class)

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling, ...]:
        assert self.max_cells is not None
        if len(comb_class.active_cells) <= self.max_cells:
            cells = self.cells_to_place(comb_class)
            if cells:
                return tuple(
                    comb_class.place_point_in_cell(cell, DIR_NONE)
                    for cell in sorted(cells)
                )
            raise StrategyDoesNotApply("The tiling is just point cells!")
        raise StrategyDoesNotApply("Too many active cells.")

    def formal_step(self) -> str:
        return f"directionless point placement (<= {self.max_cells} cells)"

    def constructor(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> DivideByN:
        if children is None:
            children = self.decomposition_function(comb_class)
        return DivideByN(
            comb_class,
            children,
            -len(self.already_placed_cells(comb_class)),
            self.extra_parameters(comb_class, children),
        )

    def reverse_constructor(
        self,
        idx: int,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> ReverseDivideByN:
        if children is None:
            children = self.decomposition_function(comb_class)
        return ReverseDivideByN(
            comb_class,
            children,
            idx,
            -len(self.already_placed_cells(comb_class)),
            self.extra_parameters(comb_class, children),
        )

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm], ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    def extra_parameters(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Dict[str, str], ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        cells = self.cells_to_place(comb_class)
        algo = RequirementPlacement(comb_class, True, True)
        res = []
        for child, cell in zip(children, sorted(cells)):
            params: Dict[str, str] = {}
            mapped_assumptions = [
                child.forward_map.map_assumption(ass).avoiding(child.obstructions)
                for ass in algo.stretched_assumptions(cell)
            ]
            for ass, mapped_ass in zip(comb_class.assumptions, mapped_assumptions):
                if mapped_ass.gps:
                    params[
                        comb_class.get_assumption_parameter(ass)
                    ] = child.get_assumption_parameter(mapped_ass)
            res.append(params)
        return tuple(res)

    def __repr__(self) -> str:
        return (
            self.__class__.__name__ + f"({self.max_cells}, {self.ignore_parent}, "
            f"{self.inferrable}, {self.possibly_empty}, {self.workable})"
        )

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d["max_cells"] = self.max_cells
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "PointingStrategy":
        return cls(**d)


class DivideByK(DivideByN):
    """
    A constructor that works as disjoint union
    but divides the values by k + shift.
    """

    def __init__(
        self,
        parent: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...],
        shift: int,
        parameter: str,
        extra_parameters: Optional[Tuple[Dict[str, str], ...]] = None,
    ):
        self.parameter = parameter
        self.division_index = parent.extra_parameters.index(parameter)
        super().__init__(parent, children, shift, extra_parameters)

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        if n + self.shift <= 0:
            return self.initial_conditions[n]
        terms = DisjointUnion.get_terms(self, parent_terms, subterms, n)
        return Counter(
            {
                key: value // (key[self.division_index] + self.shift)
                if (key[self.division_index] + self.shift) != 0
                else value
                for key, value in terms.items()
            }
        )

    def __str__(self):
        return f"divide by {self.parameter}"


class ReverseDivideByK(ReverseDivideByN):
    """
    The complement version of DivideByK.
    It works as Complement, but multiplies by k + shift the original left hand side.
    """

    def __init__(
        self,
        parent: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...],
        idx: int,
        shift: int,
        parameter: str,
        extra_parameters: Optional[Tuple[Dict[str, str], ...]] = None,
    ):
        self.parameter = parameter
        self.division_index = parent.extra_parameters.index(parameter)
        super().__init__(parent, children, idx, shift, extra_parameters)

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        if n + self.shift <= 0:
            return self.initial_conditions[n]
        parent_terms_mapped: Terms = Counter()
        for param, value in subterms[0](n).items():
            if value:
                # This is the only change from complement
                K = param[self.division_index] + self.shift
                assert K >= 0
                if K == 0:
                    parent_terms_mapped[self._parent_param_map(param)] += value
                else:
                    parent_terms_mapped[self._parent_param_map(param)] += value * K
        children_terms = subterms[1:]
        for child_terms, param_map in zip(children_terms, self._children_param_maps):
            # we subtract from total
            for param, value in child_terms(n).items():
                mapped_param = self._parent_param_map(param_map(param))
                parent_terms_mapped[mapped_param] -= value
                assert parent_terms_mapped[mapped_param] >= 0
                if parent_terms_mapped[mapped_param] == 0:
                    parent_terms_mapped.pop(mapped_param)

        return parent_terms_mapped

    def __str__(self):
        return f"reverse divide by {self.parameter}"


class AssumptionPointingStrategy(PointingStrategy):
    def __init__(
        self,
        assumption: TrackingAssumption,
        ignore_parent: bool = False,
        inferrable: bool = True,
        possibly_empty: bool = True,
        workable: bool = True,
    ):
        self.assumption = assumption
        assert not isinstance(assumption, ComponentAssumption)
        self.cells = frozenset(gp.pos[0] for gp in assumption.gps)
        super().__init__(None, ignore_parent, inferrable, possibly_empty, workable)

    def cells_to_place(self, comb_class: Tiling) -> FrozenSet[Cell]:
        return super().cells_to_place(comb_class).intersection(self.cells)

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling, ...]:
        if self.assumption not in comb_class.assumptions:
            raise StrategyDoesNotApply("The assumption is not on tiling")
        cells = self.cells_to_place(comb_class)
        if cells:
            return (
                comb_class.add_obstructions(
                    [GriddedPerm.point_perm(cell) for cell in cells]
                ),
            ) + tuple(
                comb_class.place_point_in_cell(cell, DIR_NONE) for cell in sorted(cells)
            )
        raise StrategyDoesNotApply("The assumption is just point cells!")

    def formal_step(self) -> str:
        return super().formal_step() + f" in cells {set(self.cells)}"

    def constructor(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> DivideByN:
        if children is None:
            children = self.decomposition_function(comb_class)
        return DivideByK(
            comb_class,
            children,
            -len(self.cells.intersection(self.already_placed_cells(comb_class))),
            comb_class.get_assumption_parameter(self.assumption),
            self.extra_parameters(comb_class, children),
        )

    def reverse_constructor(
        self,
        idx: int,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> ReverseDivideByN:
        if children is None:
            children = self.decomposition_function(comb_class)
        return ReverseDivideByK(
            comb_class,
            children,
            idx,
            -len(self.cells.intersection(self.already_placed_cells(comb_class))),
            comb_class.get_assumption_parameter(self.assumption),
            self.extra_parameters(comb_class, children),
        )

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str], ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        empty_params = ObstructionInferralStrategy(
            [GriddedPerm.point_perm(cell) for cell in self.cells_to_place(comb_class)]
        ).extra_parameters(comb_class)
        return empty_params + super().extra_parameters(comb_class, children[1:])

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d["assumption"] = self.assumption.to_jsonable()
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "AssumptionPointingStrategy":
        if "max_cells" in d:
            d.pop("max_cells")
        assumption = TrackingAssumption.from_dict(d.pop("assumption"))
        return cls(assumption=assumption, **d)

    def __repr__(self) -> str:
        return (
            self.__class__.__name__
            + f"({repr(self.assumption)}, {self.ignore_parent}, "
            f"{self.inferrable}, {self.possibly_empty}, {self.workable})"
        )


class AssumptionPointingFactory(StrategyFactory[Tiling]):
    def __init__(self, max_cells: int = 4) -> None:
        self.max_cells = max_cells
        super().__init__()

    def __call__(self, comb_class: Tiling) -> Iterator[AssumptionPointingStrategy]:
        if len(comb_class.active_cells) <= self.max_cells:
            for assumption in comb_class.assumptions:
                if not isinstance(assumption, ComponentAssumption):
                    yield AssumptionPointingStrategy(assumption)

    def __str__(self) -> str:
        return f"assumption pointing strategy (<= {self.max_cells} cells)"

    def __repr__(self) -> str:
        return self.__class__.__name__ + f"({self.max_cells})"

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d["max_cells"] = self.max_cells
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "AssumptionPointingFactory":

        return cls(**d)


class RequirementPointingStrategy(PointingStrategy):
    def __init__(
        self,
        gps: Tuple[GriddedPerm, ...],
        indices: Tuple[int, ...],
        ignore_parent: bool = False,
        inferrable: bool = True,
        possibly_empty: bool = True,
        workable: bool = True,
    ):
        assert len(gps) == len(indices)
        self.gps = gps
        self.indices = indices
        super().__init__(None, ignore_parent, inferrable, possibly_empty, workable)

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling, ...]:
        cells = self.cells_to_place(comb_class)
        algo = RequirementPlacement(comb_class)
        if cells:
            return algo.place_point_of_req(
                self.gps, self.indices, DIR_NONE, include_not=True, cells=cells
            )
        raise StrategyDoesNotApply("The assumption is just point cells!")

    def formal_step(self) -> str:
        return super().formal_step() + f" in {self.gps} at indices {self.indices}"

    def extra_parameters(
        self: Union[
            "RequirementPointingStrategy", "RequirementAssumptionPointingStrategy"
        ],
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Dict[str, str], ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        even_index = tuple(child for idx, child in enumerate(children) if idx % 2 == 0)
        odd_index = tuple(child for idx, child in enumerate(children) if idx % 2 == 1)
        res: List[Optional[Dict[str, str]]] = [None for _ in children]
        for idx, param in enumerate(
            PointingStrategy.extra_parameters(self, comb_class, even_index)
        ):
            res[2 * idx] = param
        for idx, param in enumerate(
            PointingStrategy.extra_parameters(self, comb_class, odd_index)
        ):
            res[2 * idx + 1] = param
        cast(List[Dict[str, str]], res)
        return tuple(cast(List[Dict[str, str]], res))

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d["gps"] = [gp.to_jsonable() for gp in self.gps]
        d["indices"] = self.indices
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "RequirementPointingStrategy":
        if "max_cells" in d:
            d.pop("max_cells")
        return cls(
            tuple(GriddedPerm.from_dict(gp) for gp in d.pop("gps")),
            tuple(d.pop("indices")),
            **d,
        )

    def __repr__(self) -> str:
        return (
            self.__class__.__name__
            + f"({self.gps}, {self.indices}, {self.ignore_parent}, "
            f"{self.inferrable}, {self.possibly_empty}, {self.workable})"
        )


class RequirementAssumptionPointingStrategy(AssumptionPointingStrategy):
    def __init__(
        self,
        gps: Tuple[GriddedPerm, ...],
        indices: Tuple[int, ...],
        ignore_parent: bool = False,
        inferrable: bool = True,
        possibly_empty: bool = True,
        workable: bool = True,
    ):
        assert len(gps) == len(indices)
        self.gps = gps
        self.indices = indices
        self.cells = frozenset(gp.pos[idx] for gp, idx in zip(gps, indices))
        self.assumption = TrackingAssumption.from_cells(self.cells)

        super().__init__(
            self.assumption, ignore_parent, inferrable, possibly_empty, workable
        )

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling, ...]:
        if self.assumption not in comb_class.assumptions:
            raise StrategyDoesNotApply("The assumption is not on tiling")
        cells = self.cells_to_place(comb_class)
        algo = RequirementPlacement(comb_class)
        if cells:
            return (
                comb_class.add_obstructions(
                    [GriddedPerm.point_perm(cell) for cell in cells]
                ),
            ) + algo.place_point_of_req(
                self.gps, self.indices, DIR_NONE, include_not=True
            )
        raise StrategyDoesNotApply("The assumption is just point cells!")

    def formal_step(self) -> str:
        return super().formal_step() + f" in {self.gps} at indices {self.indices}"

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str], ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        empty_params = ObstructionInferralStrategy(
            [GriddedPerm.point_perm(cell) for cell in self.cells_to_place(comb_class)]
        ).extra_parameters(comb_class)
        rest = RequirementPointingStrategy.extra_parameters(
            self, comb_class, children[1:]
        )
        return empty_params + tuple(rest)

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d["gps"] = [gp.to_jsonable() for gp in self.gps]
        d["indices"] = self.indices
        d.pop("assumption")
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "RequirementAssumptionPointingStrategy":
        if "max_cells" in d:
            d.pop("max_cells")
        return cls(
            tuple(GriddedPerm.from_dict(gp) for gp in d.pop("gps")),
            tuple(d.pop("indices")),
            **d,
        )

    def __repr__(self) -> str:
        return (
            self.__class__.__name__
            + f"({self.gps}, {self.indices}, {self.ignore_parent}, "
            f"{self.inferrable}, {self.possibly_empty}, {self.workable})"
        )


class RequirementPointingFactory(StrategyFactory[Tiling]):
    def __init__(self, max_cells: int = 4) -> None:
        self.max_cells = max_cells
        super().__init__()

    def __call__(self, comb_class: Tiling) -> Iterator[Rule]:
        for gps in comb_class.requirements:
            for indices in product(*[range(len(gp)) for gp in gps]):
                untracked_strategy = RequirementPointingStrategy(gps, indices)
                if len(comb_class.active_cells) <= self.max_cells:
                    yield untracked_strategy(comb_class)
                strategy = RequirementAssumptionPointingStrategy(gps, indices)
                cells_to_place = strategy.cells_to_place(comb_class)
                if (
                    untracked_strategy.cells_to_place(comb_class) != cells_to_place
                    and 0 < len(cells_to_place) <= self.max_cells
                ):
                    parent = comb_class
                    if strategy.assumption not in comb_class.assumptions:
                        rule = AddAssumptionsStrategy([strategy.assumption])(comb_class)
                        yield rule
                        parent = rule.children[0]
                    yield strategy(parent)

    def __str__(self) -> str:
        return f"requirement pointing strategy (<= {self.max_cells} cells)"

    def __repr__(self) -> str:
        return self.__class__.__name__ + f"({self.max_cells})"

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d["max_cells"] = self.max_cells
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "RequirementPointingFactory":
        return cls(**d)



from typing import TYPE_CHECKING, Dict, Tuple

from tilings.exception import InvalidOperationError

if TYPE_CHECKING:
    from tilings.assumptions import TrackingAssumption
    from tilings.griddedperm import GriddedPerm

Cell = Tuple[int, int]


class RowColMap:
    """
    A class to combine a row and a column map together and map different object related
    to tiling in accordance to those row and columns map.

    INPUT:
      - `row_map`: the row map given as a dictionary.
      - `col_map`: the column map given as a dictionary.
      - `is_identity`: A boolean that indicate if the map is the identity.
    """

    def __init__(
        self, row_map: Dict[int, int], col_map: Dict[int, int], is_identity: bool
    ) -> None:
        self._row_map = row_map
        self._col_map = col_map
        self._is_identity = is_identity

    @classmethod
    def identity(cls, dimensions: Tuple[int, int]) -> "RowColMap":
        """
        Build a map that is the identity for a tiling of the given dimensions.

        If one of the dimensions is 0 then the corresponding row/column map will
        be an empty dictionary.
        """
        col_map = {i: i for i in range(dimensions[0])}
        row_map = {i: i for i in range(dimensions[1])}
        return RowColMap(row_map=row_map, col_map=col_map, is_identity=True)

    def reverse(self) -> "RowColMap":
        """
        Return the reverse map if possible.
        Otherwise raise an InvalidOperationError.
        """
        row_map = {v: k for k, v in self._row_map.items()}
        col_map = {v: k for k, v in self._col_map.items()}
        if len(row_map) != len(self._row_map) or len(col_map) != len(self._col_map):
            raise InvalidOperationError("The map is not reversible.")
        return RowColMap(
            row_map=row_map, col_map=col_map, is_identity=self._is_identity
        )

    def is_identity(self) -> bool:
        """
        Indicate if the map is the identity map.
        """
        return self._is_identity

    def is_mappable_gp(self, gp: "GriddedPerm") -> bool:
        """
        Return True if all the cell used by the gridded perm can be mapped.
        """
        return all(self.is_mappable_cell(cell) for cell in gp.pos)

    def map_gp(self, gp: "GriddedPerm") -> "GriddedPerm":
        """
        Map the gridded permutation according to the map.
        """
        return gp.__class__(gp.patt, map(self.map_cell, gp.pos))

    def map_assumption(self, assumption: "TrackingAssumption") -> "TrackingAssumption":
        """
        Map the assumption according to the map.

        If some of the gridded permutation tracked by the assumption cannot be mapped
        they are removed from the assumption.
        """
        gps = tuple(self.map_gp(gp) for gp in assumption.gps if self.is_mappable_gp(gp))
        return assumption.__class__(gps)

    def is_mappable_cell(self, cell: Cell) -> bool:
        """
        Return True if the cell can be mapped, i.e. if the image of the row
        and the column of the are defined by the map.
        """
        return self.is_mappable_col(cell[0]) and self.is_mappable_row(cell[1])

    def map_cell(self, cell: Cell) -> Cell:
        """
        Map the cell according to the map.
        """
        return (self.map_col(cell[0]), self.map_row(cell[1]))

    def is_mappable_row(self, row: int) -> bool:
        """
        Return True if the image of the row is defined.
        """
        return row in self._row_map

    def map_row(self, row: int) -> int:
        """
        Map the row according to the map.
        """
        return self._row_map[row]

    def is_mappable_col(self, col: int) -> bool:
        """
        Return True if the image of the column is defined.
        """
        return col in self._col_map

    def map_col(self, col: int) -> int:
        """
        Map the column according to the map.
        """
        return self._col_map[col]

    def max_row(self) -> int:
        """Return the biggest row index in the image."""
        return max(self._row_map.values())

    def max_col(self) -> int:
        """Return the biggest column index in the image."""
        return max(self._col_map.values())

    def __str__(self) -> str:
        s = "RowColMap\n"
        s += f"    row map: {self._row_map}\n"
        s += f"    col map: {self._col_map}\n"
        return s



from tilings.assumptions import SkewComponentAssumption, TrackingAssumption
from tilings.griddedperm import GriddedPerm
from tilings.tilescope import TrackedClassDB
from tilings.tiling import Tiling


def test_tracked_classdb():
    tiling = Tiling(
        obstructions=(
            GriddedPerm((1, 0, 2), ((1, 0), (1, 0), (1, 0))),
            GriddedPerm((0, 2, 1, 3), ((0, 0), (0, 0), (0, 0), (0, 0))),
            GriddedPerm((0, 2, 1, 3), ((0, 0), (0, 0), (0, 0), (1, 0))),
            GriddedPerm((0, 2, 1, 3), ((0, 0), (0, 0), (1, 0), (1, 0))),
        ),
        requirements=(),
        assumptions=(
            SkewComponentAssumption((GriddedPerm((0,), ((1, 0),)),)),
            TrackingAssumption((GriddedPerm((0,), ((1, 0),)),)),
        ),
    )
    tracked_classdb = TrackedClassDB()
    tracked_classdb.add(tiling)
    new_tiling = tracked_classdb.get_class(0)
    assert tiling == new_tiling



"""The deflation strategy."""
from typing import Callable, Dict, Iterator, List, Optional, Tuple, cast

import sympy

from comb_spec_searcher import Constructor, Strategy, StrategyFactory
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.typing import (
    Parameters,
    RelianceProfile,
    SubObjects,
    SubRecs,
    SubSamplers,
    SubTerms,
    Terms,
)
from permuta import Perm
from tilings import GriddedPerm, Tiling
from tilings.assumptions import ComponentAssumption, TrackingAssumption

Cell = Tuple[int, int]


class DeflationConstructor(Constructor):
    def __init__(self, parameter: str):
        self.parameter = parameter

    def get_equation(
        self, lhs_func: sympy.Function, rhs_funcs: Tuple[sympy.Function, ...]
    ) -> sympy.Eq:
        raise NotImplementedError

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        raise NotImplementedError

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        raise NotImplementedError

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[Tuple[Parameters, Tuple[List[Optional[GriddedPerm]]]]]:
        raise NotImplementedError

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ) -> Tuple[GriddedPerm]:
        raise NotImplementedError

    def equiv(
        self, other: "Constructor", data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        raise NotImplementedError


class DeflationStrategy(Strategy[Tiling, GriddedPerm]):
    def __init__(
        self,
        cell: Cell,
        sum_deflate: bool,
        tracked: bool = True,
    ):
        self.cell = cell
        self.tracked = tracked
        self.sum_deflate = sum_deflate
        super().__init__()

    def can_be_equivalent(self) -> bool:
        return False

    def is_two_way(self, comb_class: Tiling) -> bool:
        return False

    def is_reversible(self, comb_class: Tiling) -> bool:
        return False

    def shifts(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[int, ...]:
        return (0, 0)

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling, Tiling]:
        deflated_tiling = self.deflated_tiling(comb_class)
        local_basis = comb_class.sub_tiling([self.cell])
        if self.tracked:
            return (
                deflated_tiling.add_assumption(
                    TrackingAssumption.from_cells([self.cell])
                ),
                local_basis,
            )
        return deflated_tiling, local_basis

    def deflated_tiling(self, tiling: Tiling) -> Tiling:
        if self.sum_deflate:
            extra = Perm((1, 0))
        else:
            extra = Perm((0, 1))
        reduced_reqs = tuple(
            req
            for req in tiling.requirements
            if not all(gp.pos[0] == self.cell and gp.is_single_cell() for gp in req)
        )
        reduced_ass = sorted(
            ass
            for ass in tiling.assumptions
            if not isinstance(ass, ComponentAssumption)
            or GriddedPerm.point_perm(self.cell) not in ass.gps
        )
        return Tiling(
            tiling.obstructions,
            reduced_reqs,
            reduced_ass,
            remove_empty_rows_and_cols=False,
            derive_empty=False,
            simplify=False,
            sorted_input=True,
        ).add_obstruction(extra, (self.cell, self.cell))

    def constructor(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> DeflationConstructor:
        if not self.tracked:
            raise NotImplementedError("The deflation strategy was not tracked")
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Can't deflate the cell")
        ass = TrackingAssumption.from_cells([self.cell])
        child_param = children[0].get_assumption_parameter(ass)
        gp = GriddedPerm.point_perm(self.cell)
        if any(gp in assumption.gps for assumption in comb_class.assumptions):
            raise NotImplementedError
        return DeflationConstructor(child_param)

    def reverse_constructor(
        self,
        idx: int,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Constructor:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Can't deflate the cell")
        raise NotImplementedError

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str]]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        raise NotImplementedError

    def formal_step(self) -> str:
        return f"deflating cell {self.cell}"

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm], ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d.pop("ignore_parent")
        d.pop("inferrable")
        d.pop("possibly_empty")
        d.pop("workable")
        d["cell"] = self.cell
        d["tracked"] = self.tracked
        d["sum_deflate"] = self.sum_deflate
        return d

    def __repr__(self) -> str:
        args = ", ".join(
            [
                f"cell={self.cell!r}",
                f"sum_deflate={self.sum_deflate!r}",
                f"tracked={self.tracked!r}",
            ]
        )
        return f"{self.__class__.__name__}({args})"

    @classmethod
    def from_dict(cls, d: dict) -> "DeflationStrategy":
        cell = cast(Tuple[int, int], tuple(d.pop("cell")))
        tracked = d.pop("tracked")
        sum_deflate = d.pop("sum_deflate")
        assert not d
        return cls(cell, sum_deflate, tracked)

    @staticmethod
    def get_eq_symbol() -> str:
        return "↣"


class DeflationFactory(StrategyFactory[Tiling]):
    def __init__(self, tracked: bool):
        self.tracked = tracked
        super().__init__()

    def __call__(self, comb_class: Tiling) -> Iterator[DeflationStrategy]:
        for cell in comb_class.active_cells:
            if not self._can_deflate_requirements_and_assumptions(comb_class, cell):
                continue
            if self.can_deflate(comb_class, cell, True):
                yield DeflationStrategy(cell, True, self.tracked)
            if self.can_deflate(comb_class, cell, False):
                yield DeflationStrategy(cell, False, self.tracked)

    @staticmethod
    def _can_deflate_requirements_and_assumptions(tiling: Tiling, cell: Cell) -> bool:
        def can_deflate_req_list(req: Tuple[GriddedPerm, ...]) -> bool:
            return all(gp.pos[0] == cell and gp.is_single_cell() for gp in req) or all(
                len(list(gp.points_in_cell(cell))) < 2 for gp in req
            )

        return all(can_deflate_req_list(req) for req in tiling.requirements) and all(
            not isinstance(ass, ComponentAssumption)
            or GriddedPerm.point_perm(cell) not in ass.gps
            or len(ass.gps) == 1
            for ass in tiling.assumptions
        )

    @staticmethod
    def can_deflate(tiling: Tiling, cell: Cell, sum_deflate: bool) -> bool:
        alone_in_row = tiling.only_cell_in_row(cell)
        alone_in_col = tiling.only_cell_in_col(cell)

        if alone_in_row and alone_in_col:
            return False

        deflate_patt = GriddedPerm.single_cell(
            Perm((1, 0)) if sum_deflate else Perm((0, 1)), cell
        )

        # we must be sure that no cell in a row or column can interleave
        # with any reinflated components, so collect cells that do not.
        cells_not_interleaving = set([cell])

        for ob in tiling.obstructions:
            if ob == deflate_patt:
                break  # False
            if ob.is_single_cell() or not ob.occupies(cell):
                continue
            number_points_in_cell = sum(1 for c in ob.pos if c == cell)
            if number_points_in_cell == 1:
                if len(ob) == 2:
                    # not interleaving with cell as separating if
                    # in same row or column
                    other_cell = [c for c in ob.pos if c != cell][0]
                    cells_not_interleaving.add(other_cell)
            elif number_points_in_cell == 2:
                if len(ob) != 3:
                    break  # False
                patt_in_cell = ob.get_gridded_perm_in_cells((cell,))
                if patt_in_cell != deflate_patt:
                    # you can interleave with components
                    break  # False
                # we need the other cell to be in between the intended deflate
                # patt in either the row or column
                other_cell = [c for c in ob.pos if c != cell][0]
                if (  # in a row or column
                    cell[0] == other_cell[0] or cell[1] == other_cell[1]
                ) and DeflationFactory.point_in_between(ob, cell, other_cell):
                    # this cell does not interleave with inflated components
                    cells_not_interleaving.add(other_cell)
                    continue
                break  # False
            elif number_points_in_cell >= 3:
                # you can interleave with components
                break  # False
        else:
            # check that do not interleave with any cells in row or column.
            return cells_not_interleaving >= tiling.cells_in_row(
                cell[1]
            ) and cells_not_interleaving >= tiling.cells_in_col(cell[0])
        return False

    @staticmethod
    def point_in_between(ob: GriddedPerm, cell: Cell, other_cell: Cell) -> bool:
        """Return true if point in other cell is in between point in cell.
        Assumes a length 3 pattern, and it is in a row or column."""
        row = cell[1] == other_cell[1]
        patt = cast(Tuple[int, int, int], ob.patt)
        if row:
            left = other_cell[0] < cell[0]
            if left:
                return bool(patt[0] == 1)
            return bool(patt[2] == 1)
        assert cell[0] == other_cell[0]
        below = other_cell[1] < cell[1]
        if below:
            return bool(patt[1] == 0)
        return bool(patt[1] == 2)

    def __repr__(self) -> str:
        return self.__class__.__name__ + f"({self.tracked})"

    def __str__(self) -> str:
        return ("tracked " if self.tracked else "") + "deflation factory"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["tracked"] = self.tracked
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "DeflationFactory":
        return cls(d["tracked"])



from itertools import chain, filterfalse
from typing import TYPE_CHECKING, Dict, FrozenSet, Iterable, List, Optional, Tuple

from permuta.misc import DIR_EAST, DIR_NONE, DIR_NORTH, DIR_SOUTH, DIR_WEST, DIRS
from tilings import GriddedPerm
from tilings.assumptions import TrackingAssumption

if TYPE_CHECKING:
    from tilings import Tiling

Cell = Tuple[int, int]
Dir = int
ListRequirement = List[GriddedPerm]
ObsCache = Dict[Cell, List[GriddedPerm]]
ReqsCache = Dict[Cell, List[ListRequirement]]
AssCache = Dict[Cell, List[TrackingAssumption]]


class RequirementPlacement:
    """
    The requirement placement container class.

    Places points onto own row, own col, or both.

    INPUTS:
        - `tiling`: The tilings to perform the placement with
        - `own_row`: Indiciate to place the point on its own row
        - `own_col`: Indiciate to place the point on its own column
        - `dirs`: The directions used for placement (default to all
          directions).
          The possible directions are:
            - `permuta.misc.DIR_NORTH`
            - `permuta.misc.DIR_SOUTH`
            - `permuta.misc.DIR_EAST`
            - `permuta.misc.DIR_WEST`
    """

    def __init__(
        self,
        tiling: "Tiling",
        own_row: bool = True,
        own_col: bool = True,
        dirs: Iterable[int] = tuple(DIRS),
    ):
        if not own_row and not own_col:
            raise ValueError("Must place on own row or on own column.")
        assert all(d in DIRS for d in dirs), "Got an invalid direction"
        self._tiling = tiling
        self._point_row_cells = self._tiling_point_row_cells()
        self._point_col_cells = self._tiling_point_col_cells()
        self.own_row = own_row
        self.own_col = own_col
        self._stretched_obstructions_cache: ObsCache = {}
        self._stretched_requirements_cache: ReqsCache = {}
        self._stretched_assumptions_cache: AssCache = {}
        if self.own_row and self.own_col:
            self.directions = frozenset(DIRS)
        elif self.own_row:
            self.directions = frozenset((DIR_NORTH, DIR_SOUTH))
        elif self.own_col:
            self.directions = frozenset((DIR_EAST, DIR_WEST))
        self.directions = frozenset(dirs).intersection(self.directions)
        assert self.directions, "No direction to place"

    def _tiling_point_col_cells(self) -> FrozenSet[Cell]:
        """
        The point cells of the tilings that are the only active cell in there
        column.
        """
        return frozenset(
            filter(self._tiling.only_cell_in_col, self._tiling.point_cells)
        )

    def _tiling_point_row_cells(self) -> FrozenSet[Cell]:
        """
        The point cells of the tilings that are the only active cell in there
        row.
        """
        return frozenset(
            filter(self._tiling.only_cell_in_row, self._tiling.point_cells)
        )

    def already_placed(
        self, gps: Iterable[GriddedPerm], indices: Iterable[int]
    ) -> bool:
        """
        Determine if this gps is already placed.
        """
        cells = set(gp.pos[idx] for gp, idx in zip(gps, indices))
        if len(cells) != 1:
            return False
        cell = cells.pop()
        full_placement = self.own_row and self.own_col
        if full_placement:
            return cell in self._point_col_cells and cell in self._point_row_cells
        if self.own_row:  # Only placing in own row
            return cell in self._point_row_cells
        if self.own_col:  # Only placing in own column
            return cell in self._point_col_cells
        raise Exception("Not placing at all!!")

    def _point_translation(
        self, gp: GriddedPerm, index: int, placed_cell: Cell
    ) -> Cell:
        """
        Return the translated position of the cell at the given index.

        The translation assumes that there has been a point placed in the
        position (i, j) = placed_cell where this corresponds to the index and
        value within the pattern of the gridded permutation gp.

        If the newly placed point is assumed to put on the the new column we
        have that the cell is expanded like:
            -      - - -
           | | -> | |o| |
            -      - - -
        meaning that indices to the right of i are shifted by 2.
        Similarly, for new rows we have
                   -
                  | |
            -      -
           | | -> |o|
            -      -
                  | |
                   -
        meaning that values above j are shifted by 2.
        """
        x, y = gp.pos[index]
        return (
            x + 2 if self.own_col and index >= placed_cell[0] else x,
            y + 2 if (self.own_row and gp.patt[index] >= placed_cell[1]) else y,
        )

    def _gridded_perm_translation(
        self, gp: GriddedPerm, placed_cell: Cell
    ) -> GriddedPerm:
        """
        Return the gridded permutation with all of the cells translated
        assuming that the point was placed at placed cell
        """
        newpos = [
            self._point_translation(gp, index, placed_cell) for index in range(len(gp))
        ]
        return gp.__class__(gp.patt, newpos)

    def _gridded_perm_translation_with_point(
        self, gp: GriddedPerm, point_index: int
    ) -> GriddedPerm:
        """
        Return the stretched gridded permutation obtained when the point at
        point_index in gp is placed.
        """
        # TODO: to prepare for intervals consider all ways of drawing a
        #       rectangle around point in cell.
        new_pos = [
            self._point_translation(gp, i, (point_index, gp.patt[point_index]))
            if i != point_index
            else self._placed_cell(gp.pos[point_index])
            for i in range(len(gp))
        ]
        return gp.__class__(gp.patt, new_pos)

    def _placed_cell(self, cell: Cell) -> Cell:
        """
        Return the cell in which the point will be added in the placed tiling.

        If placed on its own row, then the y coordinate is shifted by 1.
        If placed on its own column, then the x coordinate is shifted by 1.
        """
        x, y = cell
        return (x + 1 if self.own_col else x, y + 1 if self.own_row else y)

    def _point_obstructions(self, cell: Cell) -> List[GriddedPerm]:
        """
        Return the localised 12 and 21 obstruction required to ensure the
        newly placed point is a point.
        """
        placed_cell = self._placed_cell(cell)
        return [
            GriddedPerm((0, 1), (placed_cell, placed_cell)),
            GriddedPerm((1, 0), (placed_cell, placed_cell)),
        ]

    def _point_requirements(self, cell: Cell) -> List[ListRequirement]:
        """
        Return the requirement required to ensure that the newly placed point
        is a point.
        """
        placed_cell = self._placed_cell(cell)
        return [[GriddedPerm((0,), (placed_cell,))]]

    def _stretch_gridded_perm(
        self, gp: GriddedPerm, cell: Cell
    ) -> Iterable[GriddedPerm]:
        """
        Return all of the possible ways that a gridded permutation can be
        stretched assuming that a point is placed into the given cell.
        """
        mindex, maxdex, minval, maxval = gp.get_bounding_box(cell)
        if not self.own_col:
            maxdex = mindex
        elif not self.own_row:
            maxval = minval
        res = [
            self._gridded_perm_translation(gp, (i, j))
            for i in range(mindex, maxdex + 1)
            for j in range(minval, maxval + 1)
        ]
        for i in gp.points_in_cell(cell):
            res.append(self._gridded_perm_translation_with_point(gp, i))
        return res

    def _stretch_gridded_perms(
        self, gps: Iterable[GriddedPerm], cell: Cell
    ) -> List[GriddedPerm]:
        """
        Return all stretched gridded permuations for an iterable of gridded
        permutations, assuming a point is placed in the given cell.
        """
        return list(
            chain.from_iterable(self._stretch_gridded_perm(gp, cell) for gp in gps)
        )

    def stretched_obstructions(self, cell: Cell) -> List[GriddedPerm]:
        """
        Return all of the stretched obstructions that are created if placing a
        point in the given cell.
        """
        if cell not in self._stretched_obstructions_cache:
            self._stretched_obstructions_cache[cell] = self._stretch_gridded_perms(
                self._tiling.obstructions, cell
            )
        return self._stretched_obstructions_cache[cell]

    def stretched_requirements(self, cell: Cell) -> List[ListRequirement]:
        """
        Return all of the stretched requirements that are created if placing a
        point in the given cell.
        """
        if cell not in self._stretched_requirements_cache:
            self._stretched_requirements_cache[cell] = [
                self._stretch_gridded_perms(req_list, cell)
                for req_list in self._tiling.requirements
            ]
        return self._stretched_requirements_cache[cell]

    def stretched_assumptions(self, cell: Cell) -> List[TrackingAssumption]:
        """
        Return all of the stretched assumptions that are created if placing a
        point in the given cell.
        """
        if cell not in self._stretched_assumptions_cache:
            self._stretched_assumptions_cache[cell] = [
                ass.__class__(self._stretch_gridded_perms(ass.gps, cell))
                for ass in self._tiling.assumptions
            ]
        return self._stretched_assumptions_cache[cell]

    def _stretched_obstructions_requirements_and_assumptions(
        self, cell: Cell
    ) -> Tuple[List[GriddedPerm], List[ListRequirement], List[TrackingAssumption]]:
        """
        Return all of the stretched obstruction and requirements assuming that
        a point is placed in cell.
        """
        stretched_obs = self.stretched_obstructions(cell)
        stretched_reqs = self.stretched_requirements(cell)
        stretched_ass = self.stretched_assumptions(cell)
        point_obs = self._point_obstructions(cell)
        point_req = self._point_requirements(cell)
        return stretched_obs + point_obs, stretched_reqs + point_req, stretched_ass

    @staticmethod
    def _farther(c1: Cell, c2: Cell, direction: Dir) -> bool:
        """Return True if c1 is farther in the given direction than c2."""
        if direction == DIR_EAST:
            return c1[0] > c2[0]
        if direction == DIR_WEST:
            return c1[0] < c2[0]
        if direction == DIR_NORTH:
            return c1[1] > c2[1]
        if direction == DIR_SOUTH:
            return c1[1] < c2[1]
        raise Exception("Invalid direction")

    def forced_obstructions_from_requirement(
        self,
        gps: Iterable[GriddedPerm],
        indices: Iterable[int],
        cell: Cell,
        direction: Dir,
    ) -> List[GriddedPerm]:
        """
        Return the obstructions required to ensure that the placed point is
        the direction most occurence of a point used in an occurrence of any
        gridded permutation in gp_list.

        In particular, this returns the list of obstructions that are stretched
        from any gridded permutation in gp_list in which the point at idx is
        farther in the given direction than the placed cell.
        """
        placed_cell = self._placed_cell(cell)
        res = []
        if cell in self._tiling.point_cells:
            x, y = placed_cell
            if self.own_row:
                res.append(GriddedPerm.point_perm((x, y + 1)))
                res.append(GriddedPerm.point_perm((x, y - 1)))
            if self.own_col:
                res.append(GriddedPerm.point_perm((x + 1, y)))
                res.append(GriddedPerm.point_perm((x - 1, y)))
        for idx, gp in zip(indices, gps):
            # if cell is farther in the direction than gp[idx], then don't need
            # to avoid any of the stretched grided perms
            if not self._farther(cell, gp.pos[idx], direction):
                for stretched_gp in self._stretch_gridded_perm(gp, cell):
                    if self._farther(stretched_gp.pos[idx], placed_cell, direction):
                        res.append(stretched_gp)
        return res

    def _remaining_requirement_from_requirement(
        self, gps: Iterable[GriddedPerm], indices: Iterable[int], cell: Cell
    ) -> List[GriddedPerm]:
        """
        Return the requirements required to ensure that the placed point can be
        extended to be direction most occurrece of a point at the index of the
        gp in gps.

        In particulur, this returns the requirements that come from stretching
        a gridded permutation in gps, such that the point at idx is the placed
        cell.
        """
        placed_cell = self._placed_cell(cell)
        res = []
        for idx, gp in zip(indices, gps):
            if gp.pos[idx] == cell:
                for stretched_gp in self._stretch_gridded_perm(gp, cell):
                    if stretched_gp.pos[idx] == placed_cell:
                        res.append(stretched_gp)
        return res

    def place_point_of_gridded_permutation(
        self, gp: GriddedPerm, idx: int, direction: Dir
    ) -> "Tiling":
        """
        Return the tiling where the placed point correspond to the
        directionmost (the furtest in the given direction, ex: leftmost point)
        occurrence of the idx point in gp.
        """
        return self.place_point_of_req((gp,), (idx,), direction)[0]

    def place_point_of_req(
        self,
        gps: Iterable[GriddedPerm],
        indices: Iterable[int],
        direction: Dir,
        include_not: bool = False,
        cells: Optional[Iterable[Cell]] = None,
    ) -> Tuple["Tiling", ...]:
        """
        Return the tilings, where the placed point corresponds to the directionmost
        (the furtest in the given direction, ex: leftmost point) of an occurrence
        of any point idx, gp(idx) for gridded perms in gp, and idx in indices
        """
        if cells is not None:
            cells = frozenset(cells)
        else:
            cells = frozenset(gp.pos[idx] for idx, gp in zip(indices, gps))
        res = []
        for cell in sorted(cells):
            stretched = self._stretched_obstructions_requirements_and_assumptions(cell)
            (obs, reqs, ass) = stretched
            rem_req = self._remaining_requirement_from_requirement(gps, indices, cell)

            if direction == DIR_NONE:
                res.append(self._tiling.__class__(obs, reqs + [rem_req], ass))
                if include_not:
                    res.append(self._tiling.__class__(obs + rem_req, reqs, ass))
                continue
            forced_obs = self.forced_obstructions_from_requirement(
                gps, indices, cell, direction
            )
            forced_obs = [
                o1
                for o1 in forced_obs
                if not any(o2 in o1 for o2 in filterfalse(o1.__eq__, forced_obs))
            ]
            reduced_obs = [o1 for o1 in obs if not any(o2 in o1 for o2 in forced_obs)]
            reduced_obs.extend(filterfalse(reduced_obs.__contains__, forced_obs))
            res.append(
                self._tiling.__class__(
                    reduced_obs,
                    reqs + [rem_req],
                    assumptions=ass,
                    already_minimized_obs=True,
                )
            )
        return tuple(res)

    def place_point_in_cell(self, cell: Cell, direction: Dir) -> "Tiling":
        """
        Return the tiling in which a point is placed in the given direction and
        cell.
        """
        point_req = GriddedPerm((0,), (cell,))
        return self.place_point_of_req((point_req,), (0,), direction)[0]

    def col_placement(self, index: int, direction: Dir) -> Tuple["Tiling", ...]:
        """
        Return the list corresponding the index column being placed in the
        given direction.
        """
        assert direction in (DIR_EAST, DIR_WEST)
        req = [GriddedPerm((0,), (cell,)) for cell in self._tiling.cells_in_col(index)]
        return self.place_point_of_req(req, tuple(0 for _ in req), direction)

    def row_placement(self, index: int, direction: Dir) -> Tuple["Tiling", ...]:
        """
        Return the list corresponding the index row being placed in the given
        direction.
        """
        assert direction in (DIR_NORTH, DIR_SOUTH)
        req = [GriddedPerm((0,), (cell,)) for cell in self._tiling.cells_in_row(index)]
        return self.place_point_of_req(req, tuple(0 for _ in req), direction)

    def empty_col(self, index: int) -> "Tiling":
        """
        Return the tiling in which the index row is empty.
        """
        return self._tiling.add_obstructions(
            tuple(
                GriddedPerm((0,), (cell,)) for cell in self._tiling.cells_in_col(index)
            )
        )

    def empty_row(self, index: int) -> "Tiling":
        """
        Return the tiling in which the index row is empty.
        """
        return self._tiling.add_obstructions(
            tuple(
                GriddedPerm((0,), (cell,)) for cell in self._tiling.cells_in_row(index)
            )
        )



import abc
from functools import partial
from itertools import chain, combinations
from typing import Dict, Iterable, Iterator, List, Optional, Set, Tuple, Type, cast

from comb_spec_searcher import StrategyFactory, SymmetryStrategy
from comb_spec_searcher.exception import StrategyDoesNotApply
from permuta import Perm
from tilings import GriddedPerm, Tiling
from tilings.assumptions import (
    ComponentAssumption,
    SkewComponentAssumption,
    SumComponentAssumption,
    TrackingAssumption,
)

__all__ = ("SymmetriesFactory",)


class TilingSymmetryStrategy(SymmetryStrategy[Tiling, GriddedPerm]):
    @abc.abstractmethod
    def gp_transform(self, tiling: Tiling, gp: GriddedPerm) -> GriddedPerm:
        pass

    @abc.abstractmethod
    def inverse_gp_transform(self, tiling: Tiling, gp: GriddedPerm) -> GriddedPerm:
        pass

    @staticmethod
    def assumption_type_transform(
        assumption: TrackingAssumption,
    ) -> Type[TrackingAssumption]:
        raise NotImplementedError

    @staticmethod
    def _assumption_type_swap(
        assumption: TrackingAssumption,
    ) -> Type[TrackingAssumption]:
        if isinstance(assumption, ComponentAssumption):
            if isinstance(assumption, SumComponentAssumption):
                return SkewComponentAssumption
            return SumComponentAssumption
        return assumption.__class__

    @staticmethod
    def _assumption_type_identity(
        assumption: TrackingAssumption,
    ) -> Type[TrackingAssumption]:
        return assumption.__class__

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling, ...]:
        return (
            Tiling(
                tuple(
                    map(partial(self.gp_transform, comb_class), comb_class.obstructions)
                ),
                tuple(
                    tuple(map(partial(self.gp_transform, comb_class), req))
                    for req in comb_class.requirements
                ),
                tuple(
                    self.__class__.assumption_type_transform(ass)(
                        map(partial(self.gp_transform, comb_class), ass.gps)
                    )
                    for ass in comb_class.assumptions
                ),
                remove_empty_rows_and_cols=False,
                derive_empty=False,
                simplify=False,
            ),
        )

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str], ...]:
        if not comb_class.extra_parameters:
            return super().extra_parameters(comb_class, children)
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        child = children[0]
        mapped_assumptions = tuple(
            self.__class__.assumption_type_transform(ass)(
                tuple(self.gp_transform(comb_class, gp) for gp in ass.gps)
            )
            for ass in comb_class.assumptions
        )
        return (
            {
                comb_class.get_assumption_parameter(
                    assumption
                ): child.get_assumption_parameter(mapped_assumption)
                for assumption, mapped_assumption in zip(
                    comb_class.assumptions, mapped_assumptions
                )
            },
        )

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        """This method will enable us to generate objects, and sample."""
        yield self.inverse_gp_transform(comb_class, cast(GriddedPerm, objs[0]))

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[GriddedPerm]:
        """This function will enable us to have a quick membership test."""
        return (self.gp_transform(comb_class, obj),)

    def __repr__(self) -> str:
        return self.__class__.__name__ + "()"

    @classmethod
    def from_dict(cls, d: dict) -> "TilingSymmetryStrategy":
        return cls()


class TilingReverse(TilingSymmetryStrategy):
    """
    Flips the tiling on its vertical axis.
    """

    def gp_transform(self, tiling: Tiling, gp: GriddedPerm) -> GriddedPerm:
        def reverse_cell(cell):
            return (tiling.dimensions[0] - cell[0] - 1, cell[1])

        return gp.reverse(reverse_cell)

    def inverse_gp_transform(self, tiling: Tiling, gp: GriddedPerm) -> GriddedPerm:
        return self.gp_transform(tiling, gp)

    assumption_type_transform = TilingSymmetryStrategy._assumption_type_swap

    def formal_step(self) -> str:
        return "reverse of the tiling"

    def __str__(self) -> str:
        return "reverse"


class TilingComplement(TilingSymmetryStrategy):
    """
    Flips the tiling over the horizontal axis.
    """

    def gp_transform(self, tiling: Tiling, gp: GriddedPerm) -> GriddedPerm:
        def complement_cell(cell):
            return (cell[0], tiling.dimensions[1] - cell[1] - 1)

        return gp.complement(complement_cell)

    def inverse_gp_transform(self, tiling: Tiling, gp: GriddedPerm) -> GriddedPerm:
        return self.gp_transform(tiling, gp)

    assumption_type_transform = TilingSymmetryStrategy._assumption_type_swap

    def formal_step(self) -> str:
        return "complement of the tiling"

    def __str__(self) -> str:
        return "complement"


class TilingInverse(TilingSymmetryStrategy):
    """
    Flips the tiling over the diagonal.
    """

    def gp_transform(self, tiling: Tiling, gp: GriddedPerm) -> GriddedPerm:
        def inverse_cell(cell):
            return (cell[1], cell[0])

        return gp.inverse(inverse_cell)

    def inverse_gp_transform(self, tiling: Tiling, gp: GriddedPerm) -> GriddedPerm:
        return self.gp_transform(tiling, gp)

    assumption_type_transform = TilingSymmetryStrategy._assumption_type_identity

    def formal_step(self) -> str:
        return "inverse of the tiling"

    def __str__(self) -> str:
        return "inverse"


class TilingAntidiagonal(TilingSymmetryStrategy):
    """
    Flips the tiling over the antidiagonal.
    """

    def gp_transform(self, tiling: Tiling, gp: GriddedPerm) -> GriddedPerm:
        def antidiagonal_cell(cell):
            return (
                tiling.dimensions[1] - cell[1] - 1,
                tiling.dimensions[0] - cell[0] - 1,
            )

        return gp.antidiagonal(antidiagonal_cell)

    def inverse_gp_transform(self, tiling: Tiling, gp: GriddedPerm) -> GriddedPerm:
        def antidiagonal_cell(cell):
            return (
                tiling.dimensions[0] - cell[1] - 1,
                tiling.dimensions[1] - cell[0] - 1,
            )

        return gp.antidiagonal(antidiagonal_cell)

    assumption_type_transform = TilingSymmetryStrategy._assumption_type_identity

    def formal_step(self) -> str:
        return "antidiagonal of the tiling"

    def __str__(self) -> str:
        return "antidiagonal"


class TilingRotate90(TilingSymmetryStrategy):
    """
    Rotate the tiling 90 degrees clockwise.
    """

    def gp_transform(self, tiling: Tiling, gp: GriddedPerm) -> GriddedPerm:
        def rotate90_cell(cell):
            return (cell[1], tiling.dimensions[0] - cell[0] - 1)

        return gp.rotate90(rotate90_cell)

    def inverse_gp_transform(self, tiling: Tiling, gp: GriddedPerm) -> GriddedPerm:
        def rotate270_cell(cell):
            return (tiling.dimensions[0] - cell[1] - 1, cell[0])

        return gp.rotate270(rotate270_cell)

    assumption_type_transform = TilingSymmetryStrategy._assumption_type_swap

    def formal_step(self) -> str:
        return "rotate the tiling 90 degrees clockwise"

    def __str__(self) -> str:
        return "rotate90"


class TilingRotate180(TilingSymmetryStrategy):
    """
    Rotate the tiling 180 degrees clockwise.
    """

    def gp_transform(self, tiling: Tiling, gp: GriddedPerm) -> GriddedPerm:
        def rotate180_cell(cell):
            return (
                tiling.dimensions[0] - cell[0] - 1,
                tiling.dimensions[1] - cell[1] - 1,
            )

        return gp.rotate180(rotate180_cell)

    def inverse_gp_transform(self, tiling: Tiling, gp: GriddedPerm) -> GriddedPerm:
        return self.gp_transform(tiling, gp)

    assumption_type_transform = TilingSymmetryStrategy._assumption_type_identity

    def formal_step(self) -> str:
        return "rotate the tiling 180 degrees clockwise"

    def __str__(self) -> str:
        return "rotate180"


class TilingRotate270(TilingSymmetryStrategy):
    """
    Rotate the tiling 270 degrees clockwise.
    """

    def gp_transform(self, tiling: Tiling, gp: GriddedPerm) -> GriddedPerm:
        def rotate270_cell(cell):
            return (tiling.dimensions[1] - cell[1] - 1, cell[0])

        return gp.rotate270(rotate270_cell)

    def inverse_gp_transform(self, tiling: Tiling, gp: GriddedPerm) -> GriddedPerm:
        def rotate90_cell(cell):
            return (cell[1], tiling.dimensions[1] - cell[0] - 1)

        return gp.rotate90(rotate90_cell)

    assumption_type_transform = TilingSymmetryStrategy._assumption_type_swap

    def formal_step(self) -> str:
        return "rotate the tiling 270 degrees clockwise"

    def __str__(self) -> str:
        return "rotate270"


class SymmetriesFactory(StrategyFactory[Tiling]):
    """
    Yield symmetry strategies such that all of the underlying patterns of
    obstructions of the symmetric tiling are subpatterns of the given basis.
    """

    def __init__(
        self,
        basis: Optional[Iterable[Perm]] = None,
    ):
        self._basis = tuple(basis) if basis is not None else None
        if self._basis is not None:
            assert all(
                isinstance(p, Perm) for p in self._basis
            ), "Element of the basis must be Perm"
            self.subpatterns: Set[Perm] = set(
                chain.from_iterable(self._subpatterns(p) for p in self._basis)
            )
            self.acceptablesubpatterns: List[Set[Perm]] = [
                set(p for p in self.subpatterns if p.rotate() in self.subpatterns),
                set(p for p in self.subpatterns if p.rotate(2) in self.subpatterns),
                set(p for p in self.subpatterns if p.rotate(3) in self.subpatterns),
                set(p for p in self.subpatterns if p.inverse() in self.subpatterns),
                set(p for p in self.subpatterns if p.reverse() in self.subpatterns),
                set(
                    p
                    for p in self.subpatterns
                    if p.flip_antidiagonal() in self.subpatterns
                ),
                set(p for p in self.subpatterns if p.complement() in self.subpatterns),
            ]
        super().__init__()

    def __call__(self, tiling: Tiling) -> Iterator[TilingSymmetryStrategy]:
        underlying_patts = set(gp.patt for gp in tiling.obstructions)
        if (
            self._basis is None
            or all(p in self.acceptablesubpatterns[0] for p in underlying_patts)
            or self._basis is None
        ):
            yield TilingRotate90()
        if (
            self._basis is None
            or all(p in self.acceptablesubpatterns[1] for p in underlying_patts)
            or self._basis is None
        ):
            yield TilingRotate180()
        if (
            self._basis is None
            or all(p in self.acceptablesubpatterns[2] for p in underlying_patts)
            or self._basis is None
        ):
            yield TilingRotate270()
        if (
            self._basis is None
            or all(p in self.acceptablesubpatterns[3] for p in underlying_patts)
            or self._basis is None
        ):
            yield TilingInverse()
        if (
            self._basis is None
            or all(p in self.acceptablesubpatterns[4] for p in underlying_patts)
            or self._basis is None
        ):
            yield TilingReverse()
        if (
            self._basis is None
            or all(p in self.acceptablesubpatterns[5] for p in underlying_patts)
            or self._basis is None
        ):
            yield TilingAntidiagonal()
        if (
            self._basis is None
            or all(p in self.acceptablesubpatterns[6] for p in underlying_patts)
            or self._basis is None
        ):
            yield TilingComplement()

    @staticmethod
    def _subpatterns(perm: Perm) -> Iterator[Perm]:
        for n in range(len(perm) + 1):
            for indices in combinations(range(len(perm)), n):
                yield Perm.to_standard([perm[i] for i in indices])

    def change_basis(
        self,
        basis: Iterable[Perm],
    ) -> "SymmetriesFactory":
        """
        Return the version of the strategy with the given basis instead
        of the current one.
        """
        return self.__class__(tuple(basis))

    @property
    def basis(self) -> Optional[Tuple[Perm, ...]]:
        return self._basis

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["basis"] = self._basis
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "SymmetriesFactory":
        if "basis" in d and d["basis"] is not None:
            basis: Optional[List[Perm]] = [Perm(p) for p in d.pop("basis")]
        else:
            basis = d.pop("basis", None)
        return cls(basis=basis)

    def __str__(self) -> str:
        if self._basis is not None:
            basis = ", ".join(str(p) for p in self._basis)
            return f"symmetries in Av({basis})"
        return "all symmetries"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(basis={self._basis})"



"""
This file contains "experimental" verification strategies, which are those that verify
tilings for which we are not certain we can independently calculate their counting
sequence or generating function.
"""

from typing import Iterable, Iterator, Optional, Tuple

from comb_spec_searcher import StrategyFactory, VerificationStrategy
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.strategies import VerificationRule
from permuta import Av, Perm
from tilings import GriddedPerm, Tiling
from tilings.algorithms import Factor, SubclassVerificationAlgorithm

from .abstract import BasisAwareVerificationStrategy

__all__ = [
    "NoRootCellVerificationStrategy",
    "ShortObstructionVerificationStrategy",
    "SubclassVerificationFactory",
]

TileScopeVerificationStrategy = VerificationStrategy[Tiling, GriddedPerm]


class NoRootCellVerificationStrategy(BasisAwareVerificationStrategy):
    """
    A strategy to mark as verified any tiling that does not contain the root
    basis localized in a cell. Tilings with dimensions 1x1 are ignored.
    """

    def verified(self, comb_class: Tiling):
        return comb_class.dimensions != (1, 1) and all(
            frozenset(obs) not in self.symmetries
            for obs, _ in comb_class.cell_basis().values()
        )

    def formal_step(self) -> str:
        basis = ", ".join(str(p) for p in self.basis)
        return f"tiling has no Av({basis}) cell"

    def __str__(self) -> str:
        basis = ", ".join(str(p) for p in self.basis)
        return f"no Av({basis}) cell verification"


class ShortObstructionVerificationStrategy(BasisAwareVerificationStrategy):
    """
    A strategy to mark as verified any tiling whose crossing obstructions all have
    size at most 3. Tilings with dimensions 1x1 are ignored.
    """

    def __init__(
        self,
        short_length: int = 3,
        basis: Optional[Iterable[Perm]] = None,
        symmetry: bool = False,
        ignore_parent: bool = True,
    ):
        self.short_length = short_length
        super().__init__(basis=basis, symmetry=symmetry, ignore_parent=ignore_parent)

    def verified(self, comb_class: Tiling):
        return comb_class.dimensions != (1, 1) and all(
            ob.is_single_cell() or len(ob) <= self.short_length
            for ob in comb_class.obstructions
        )

    def change_basis(
        self, basis: Iterable[Perm], symmetry: bool
    ) -> "ShortObstructionVerificationStrategy":
        """
        Return a new version of the verification strategy with the given basis instead
        of the current one.
        """
        basis = tuple(basis)
        return ShortObstructionVerificationStrategy(
            self.short_length, basis, symmetry, self.ignore_parent
        )

    def formal_step(self) -> str:
        return f"tiling has short (length <= {self.short_length}) crossing obstructions"

    def decomposition_function(
        self, comb_class: Tiling
    ) -> Optional[Tuple[Tiling, ...]]:
        """
        The rule as the root as children if one of the cell of the tiling is the root.
        """
        if self.verified(comb_class):
            if not self.basis:
                return ()
            for obs, _ in comb_class.cell_basis().values():
                if frozenset(obs) in self.symmetries:
                    return (Tiling.from_perms(self.basis),)
            return ()
        return None

    def shifts(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[int, ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply
        if children:
            return (0,)
        return ()

    def __str__(self) -> str:
        return (
            f"short (length <= {self.short_length}) crossing obstruction verification"
        )

    def __repr__(self) -> str:
        args = ", ".join(
            [
                f"basis={self._basis}",
                f"short_length={self.short_length}",
                f"ignore_parent={self.ignore_parent}",
            ]
        )
        return f"{self.__class__.__name__}({args})"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["short_length"] = self.short_length
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "ShortObstructionVerificationStrategy":
        if "basis" in d and d["basis"] is not None:
            basis: Optional[Tuple[Perm, ...]] = tuple(Perm(p) for p in d.pop("basis"))
        else:
            basis = d.pop("basis", None)
        return cls(basis=basis, **d)


class SubclassVerificationStrategy(TileScopeVerificationStrategy):
    """
    A strategy object for holding the results of SubclassVerification
    """

    def __init__(self, subclass_basis=Iterable[Perm], ignore_parent: bool = True):
        self.subclass_basis = tuple(sorted(subclass_basis))
        super().__init__(ignore_parent=ignore_parent)

    def verified(self, comb_class: Tiling) -> bool:
        algo = SubclassVerificationAlgorithm(comb_class, set(self.subclass_basis))
        return algo.subclasses == self.subclass_basis

    def formal_step(self) -> str:
        return f"tiling is contained in the subclass {Av(self.subclass_basis)}"

    def __str__(self) -> str:
        return "subclass verification strategy"

    def __repr__(self):
        args = ", ".join(
            [
                f"subclass_basis={self.subclass_basis}",
                f"ignore_parent={self.ignore_parent}",
            ]
        )
        return f"{self.__class__.__name__}({args})"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["subclass_basis"] = self.subclass_basis
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "SubclassVerificationStrategy":
        subclass_basis = [Perm(p) for p in d.pop("subclass_basis")]
        return cls(subclass_basis=subclass_basis, **d)


class SubclassVerificationFactory(StrategyFactory[Tiling]):
    """
    A factory that verifies a tiling if its underlying ungridded permutations are
    contained in a subclass of the search class, and if so creates the appropriate
    strategy.

    Note: it isn't really a generator.
    """

    def __init__(self, perms_to_check: Optional[Iterable[Perm]] = None):
        if perms_to_check is None:
            self.perms_to_check = None
        else:
            self.perms_to_check = set(perms_to_check)
        super().__init__()

    def __call__(self, comb_class: Tiling) -> Iterator[VerificationRule]:
        assert self.perms_to_check is not None, "perms_to_check was never set"

        # It is a waste of time to check a factorable tiling, since we will check its
        # children eventually.
        if Factor(comb_class).factorable():
            return

        algo = SubclassVerificationAlgorithm(comb_class, self.perms_to_check)
        if algo.subclasses:
            yield SubclassVerificationStrategy(algo.subclasses)(comb_class, tuple())

    def change_perms(
        self, perms_to_check: Iterable[Perm]
    ) -> "SubclassVerificationFactory":
        """
        Return a new version of the verfication strategy with the given perms to check
        instead of the current one.
        """
        perms_to_check = set(perms_to_check)
        return self.__class__(perms_to_check=perms_to_check)

    def __str__(self):
        return "subclass verification factory"

    def __repr__(self):
        return f"{self.__class__.__name__}(perms_to_check={self.perms_to_check})"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["perms_to_check"] = (
            tuple(sorted(self.perms_to_check))
            if self.perms_to_check is not None
            else None
        )
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "SubclassVerificationFactory":
        if d["perms_to_check"] is not None:
            perms_to_check: Optional[Iterable[Perm]] = [
                Perm(p) for p in d.pop("perms_to_check")
            ]
        else:
            perms_to_check = d.pop("perms_to_check")
        assert len(d) == 0
        return cls(perms_to_check=perms_to_check)
